// tlv: a utility package used for parsing TLV structures
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The PIV Middleware Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.
#include "tlvparse.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

/**
	This function is the generic function that uses va_args to let users pass in a single (T-L-V,T-L-V,...) buffer, or, alternatively, two (tl-buffer, v-buffer) buffers. The function can be used in one of the following forms: \n <tt>&nbsp;&nbsp;&nbsp;&nbsp;parseTLVs(ptlvs, ptlvCount, 1, buffer, bufferLength);</tt> \n or \n <tt>&nbsp;&nbsp;&nbsp;&nbsp;parseTLVs(ptlvs, ptlvCount, 2, tlBuffer, tlBufferLen, vBuffer, vBufferLen);
	
	\param tlvs A pointer to an array of tlv structures.
	\param ptlvCount The function sets this to indicate the number of TLVs that there are.
	\param bCount This tells the function how many buffers to expect.  

	\return SUCCESS \n BAD_TLEN \n BAD_VLEN \n BAD_LEN \n NOT_ENOUGH_MEMORY \n BAD_TBUFF \n UNSUPPORTED
*/
int parseTLVs(PTLV tlvs[], unsigned int *ptlvCount, unsigned int bCount, ...) {
	unsigned char *b1, *b2;
	unsigned int b1Len, b2Len;
	va_list ap;
	va_start(ap,bCount);
	if (bCount==1) {
		b1 = va_arg(ap,unsigned char*);
		b1Len = va_arg(ap,unsigned int);
		return parseTLVTLV(b1,b1Len,tlvs,ptlvCount);
	}
	if (bCount == 2) {
		b1 = va_arg(ap,unsigned char*);
		b1Len = va_arg(ap,unsigned int);
		b2 = va_arg(ap,unsigned char*);
		b2Len = va_arg(ap,unsigned int);
		return parseTBuffVBuff(b1,b1Len,b2,b2Len,tlvs,ptlvCount);
	}
	else return TLV_UNSUPPORTED;
}

/** 
	This function is the non-generic version of the ParseTLVs function.  The result is the same as an appropriatley crafted version of the parseTLVs function.
	
	\param tBuff A pointer to the tag-length buffer.
	\param tLen The size of <tt>tBuff</tt>.
	\param vBuff A pointer to the value buffer.
	\param vLen The size of <tt>vBuff</tt>.
	\param tlvs An array of pointers to TLV structures (the results of the parsing).
	\param ptlvCount An unsigned int that contains the number of TLV structures in <tt>tlvs</tt>.

	\return SUCCESS \n BAD_TLEN \n BAD_VLEN \n BAD_TBUFF \n NOT_ENOUGH_MEMORY
*/
int parseTBuffVBuff ( unsigned char *tBuff, unsigned int tLen, unsigned char *vBuff, unsigned int vLen, PTLV tlvs[], unsigned int *ptlvCount) {
	unsigned int v=0, t=0;
	unsigned int length, j;
	int i;
	unsigned char tag;
	/* an empty CCC should at have 0x0000 at the beginning */
	if (tLen<HEADER_LEN) return TLV_BAD_TLEN;
	if (vLen<HEADER_LEN) return TLV_BAD_VLEN;
#ifdef CAC
	for (i=HEADER_LEN-1; i>=0; i--) { /* Had to change this for the CAC - in general? */
		v = (v<<8) | vBuff[i];
		t = (t<<8) | tBuff[i];
	}
	if (v!=vLen-2) return TLV_BAD_VLEN; /* Have to account for the 2 length bytes as well... */
	if (t!=tLen-2) return TLV_BAD_TLEN; /* Have to account for the 2 length bytes as well... */
#else
	for (i=0; i<HEADER_LEN; i++) {
		v = (v<<8) | vBuff[i];
		t = (t<<8) | tBuff[i];
	}
	if (v!=vLen) return TLV_BAD_VLEN;
	if (t!=tLen) return TLV_BAD_TLEN;
#endif
	if ((tLen-HEADER_LEN)%2) return TLV_BAD_TBUFF;
	*ptlvCount=0;
	v=HEADER_LEN;
	for (t=HEADER_LEN; t<tLen-1;) {
		tag = tBuff[t++];
		length = getLen(tBuff,tLen,&t);
		t++;
		
		if (tlvs != NULL) {
			if(tlvs[*ptlvCount] != NULL) 
				free(tlvs[*ptlvCount]);
			else {

				if (length != 0)
					tlvs[*ptlvCount] = (TLV*)calloc((sizeof(TLV)-1)+length,1);
				else
					tlvs[*ptlvCount] = (TLV*)calloc(sizeof(TLV),1);
			}
		
			if (tlvs[*ptlvCount]==NULL) return TLV_NOT_ENOUGH_MEMORY;
			tlvs[*ptlvCount]->tag=tag;
			tlvs[*ptlvCount]->length=length;
		}
		for(j=0; j<length; j++) {
			if (v>=vLen) return TLV_BAD_VLEN;
			if (tlvs != NULL) tlvs[*ptlvCount]->value[j]=vBuff[v];
			v++;
		}
		(*ptlvCount)++;
	}
	return TLV_SUCCESS;
}

/** 
	This is another non-generic version of the ParseTLVs function.  This is for the single-buffer T-L-V,T-L-V,... structure found in File System GSC-IS cards.
	\param buffer The buffer that contains the t-l-v... structure.
	\param buffLen The size of <tt>buffer</tt>.
	\param tlvs The array of pointers to TLV structures that the results of the parsing is stored in.
	\param ptlvCount The number of TLV structures that are in <tt>tlvs</tt>.
	\return SUCCESS \n BAD_LEN \n NOT_ENOUGH_MEMORY
*/
int parseTLVTLV (unsigned char *buffer, unsigned int buffLen, PTLV tlvs[], unsigned int *ptlvCount) {
	unsigned int length=0;
	unsigned int loc=0;
	unsigned char tag;
	unsigned int i;
	if (buffLen<=2) return TLV_BAD_LEN;
		*ptlvCount=0;
	if (((int)buffer[0])==0xEE) {
		if (buffLen<=4) return TLV_BAD_LEN;
		loc = buffer[1]+2;
		for (i=loc-1; i>2; i--)
			length = length<<8 | ((int)buffer[i]);
		if (length != buffLen) return TLV_BAD_LEN;
	}
	while(loc<buffLen) {
		tag = (unsigned char)buffer[loc++];
		length = getLen(buffer, buffLen, &loc);
		loc++;
		if (tlvs!=NULL) {
			if (loc+length > buffLen) 
				return TLV_BAD_LEN;
			if(tlvs[*ptlvCount] != NULL)
				free(tlvs[*ptlvCount]);
			else {

				if (length != 0)
					tlvs[*ptlvCount] = (TLV*)calloc((sizeof(TLV)-1)+length,1);
				else
					tlvs[*ptlvCount] = (TLV*)calloc(sizeof(TLV),1);
			}

			if (tlvs[*ptlvCount]==NULL) return TLV_NOT_ENOUGH_MEMORY;
			tlvs[*ptlvCount]->tag=tag;
			tlvs[*ptlvCount]->length=length;
		}
		for (i=0; i<length; i++) {
			if (tlvs != NULL) tlvs[*ptlvCount]->value[i] = buffer[loc];
			loc++;
		}
		(*ptlvCount)++;
	}
	return TLV_SUCCESS;
}

unsigned int getLen ( unsigned char *buffer, unsigned int len, unsigned int *loc ) {
	if (len<=*loc) return -1;
	if (((unsigned int)buffer[*loc])==0xFF) {
		if (len<=(*loc)+2) return -2;
		(*loc)=(*loc)+2;
		return (((unsigned int)buffer[*loc])<<8) | ((unsigned int)buffer[(*loc)-1]);
	}
	return (unsigned int)buffer[*loc];
}
/** 
	This function intelligently frees the memory of a PTLV array.
	\param tlvs This is the array of pointers to the TLV structurs that are to be freed by this function.
	\param i The number of TLV structurs in <tt>tlvs</tt>.
	\return void
*/
void freeTLVArray(PTLV tlvs[], int i) {
	for(;--i>=0;)
		free(tlvs[i]);
}
