// PIV Card Application
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and
// Technology (NIST), an agency of the Federal Government.  Pursuant to
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
//
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY,
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
//
// Permission to use this software is contingent upon your acceptance of
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__PIV_CARD_APPLICATION_H_INCLUDED__)
#define __PIV_CARD_APPLICATION_H_INCLUDED__

#if defined(_MSC_VER )
  // It's OK that the compiler can't automatically generate default constructors or
  // assignment operators for class PIV_CARD_APPLICATION_DLLes with const data members
  // We're shooting for immutability here
  #if defined(PIV_CARD_APPLICATION_EXPORTS)
    #define PIV_CARD_APPLICATION_DLL _declspec(dllexport)
  #else //  defined ( PIV_CARD_APPLICATION_EXPORTS )
    #define PIV_CARD_APPLICATION_DLL _declspec(dllimport)
  #endif  // defined ( PIV_CARD_APPLICATION_EXPORTS )
#else
   #define PIV_CARD_APPLICATION_DLL
#endif

#include <File_System.h>
#include <Application.h>
#include <Key_Generation.h>

// Table 8 - Algorithm Identifiers
#define DEFAULT_ALGORITHM   0x00 // The default PIV algorithm is TRIPLE_DES_ECB_192
#define TRIPLE_DES_ECB_128  0x01
#define TRIPLE_DES_CBC_128  0x02
#define TRIPLE_DES_ECB_192  0x03
#define TRIPLE_DES_CBC_192  0x04
#define RSA_3072            0x05
#define RSA_1024            0x06
#define RSA_2048            0x07
#define AES_ECB_128         0x08
#define AES_CBC_128         0x09
#define AES_ECB_192         0x0A
#define AES_CBC_192         0x0B
#define AES_ECB_256         0x0C
#define AES_CBC_256         0x0D
#define ECC_P_224           0x0E
#define ECC_K_233           0x0F
#define ECC_B_233           0x10
#define ECC_P_256           0x11
#define ECC_K_283           0x12
#define ECC_B_283           0x13

#define DEFAULT_ALGORITHM_IDENTIFIER    TRIPLE_DES_ECB_192

// Table 13 - Key References
#define CARDHOLDER_GLOBAL_PIN                           0x00    // 8-Byte PIN
#define CARDHOLDER_PIV_CARD_APPLICATION_PIN             0x80    // 8-Byte PIN
#define CARDHOLDER_PIV_CARD_APPLICATION_PUK             0x81    // 8-Byte PIN Unblocking Key
#define PIV_CARD_APPLICATION_VALIDATION                 0x9A    // 1024-Bit RSA
#define PIV_CARD_APPLICATION_ADMINISTRATION             0x9B    // 192-Byte Triple DES ECB
#define PIV_CARD_APPLICATION_DIGITAL_SIGNATURE          0x9C    // 1024-Bit RSA
#define PIV_CARD_APPLICATION_KEY_MANAGEMENT             0x9D    // 1024-Bit RSA
#define PIV_CARD_AUTHENTICATION_KEY                     0x9E    // Unspecified
#define DUMMY_KEY_REFERENCE								0x01    // Dummy key reference used for initializing

// Fixed PIV Card Application Parameters
#define MANDATORY_PIN_LENGTH                            0x08    // Right-padded with 0xFF
#define MANDATORY_CHALLENGE_LENGTH                      0x08
#define MAXIMUM_CRYPTOGRAM_LENGTH                      0x180
#define MAXIMUM_RESPONSE							   0x400
#define MAXIMUM_GET_DATA                               12750    // Maximum data object size + a little for comfort
#define RSA_ENCRYPTED_CHALLENGE_LENGTH                  0x80

#define CHAIN_BIT_MASK                                0xEFFF

static Octet get_challenge_response_octets [ ]
    = { DYN_TEMPLATE, 0x0A, DYN_CHALLENGE, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } ;

static Octet internal_authenticate_response_RSA_1024 [134]
    = { DYN_TEMPLATE, 0x81, 0x83, DYN_RESPONSE,  0x81, 0x80 } ;
static Octet internal_authenticate_response [ ]
    = { DYN_TEMPLATE, 0x0A, DYN_RESPONSE,  0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } ;

static Octet mutual_authenticate_response_RSA_1024 [134]
    = { DYN_TEMPLATE, 0x81, 0x83, DYN_WITNESS,  0x81, 0x80 } ;
static Octet mutual_authenticate_response [ ]
    = { DYN_TEMPLATE, 0x0A, DYN_WITNESS,  0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } ;

static Octet the_digital_signature_certificate_tag [ ] = { 0x5F, 0xC1, 0x0A } ;
static Octet PIN_Always = 0 ;
static bool mutual_authentication_passed = false;
static bool external_authentication_passed = false;

static unsigned int key_size_in_bits [ ]
= { 0 , 0 , 0 , 0 , 0 , 3072 , 1024 , 2048 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 };
static Octet publicKeyTemplate [ ] = { PK_TEMPLATE } ;
static Octet dataTemplate [ ] = { DATA_FIELD_DATA , 0 } ;

static Octet PIVCardApplicationProperties [ ] = {
    APT_TEMPLATE, 0x6C,
                  APT_AID, 0x0B, 0xA0, 0x00, 0x00, 0x03, 0x08, 0x00, 0x00, 0x10, 0x00, 0x01, 0x00,
                  APT_TAG_AUTHORITY, 0x07, APT_TAG_AID, 0x05, 0xA0, 0x00, 0x00, 0x03, 0x08,
                  APT_APPLICATION_LABEL, 0x14,
                                 'P','I','V',' ','C','a','r','d',' ','A','p','p','l','i','c','a','t','i','o','n',
                  APT_URL, 0x3D, 'c','s','r','c','.','n','i','s','t','.','g','o','v','/',
                                 'p','u','b','l','i','c','a','t','i','o','n','s','/',
                                 'n','i','s','t','p','u','b','s','/','8','0','0','-','7','3','/',
                                 'S','P','8','0','0','-','7','3','-','F','i','n','a','l','.','p','d','f'};

class PIV_CARD_APPLICATION_DLL PIV_Card_Application : public Application
{

public:
  PIV_Card_Application( File_Manager * file_manager_pointer ,
                           const File_System_Pointer & file_system_pointer ,
                           const Security_System_Pointer & security_system_pointer ,
                           Octet_String_Pointer & challenge_pointer ) ;

  virtual long exchange_APDUs   ( const CAPDU_Packet & command ,
                                        RAPDU_Packet & response ) ;

  virtual const Octet_String & AID ( ) { return the_AID ; }

private:

    typedef struct { CLAINS clains ; APDU_Handler PIV_Card_Application::* APDU_handler ; } Dispatch_Table_Entry ;
    static Dispatch_Table_Entry dispatch_table [ ] ;

    APDU_Handler select_application             ;
    APDU_Handler get_data                       ;
    APDU_Handler verify_pin                     ;
    APDU_Handler change_reference_data          ;
    APDU_Handler reset_retry_counter            ;
    APDU_Handler general_authenticate           ;
    APDU_Handler put_data                       ;
    APDU_Handler generate_asymmetric_key_pair   ;
    APDU_Handler command_chaining               ;

    Octet_String the_AID                        ;

    CLAINS the_chained_command                  ;
    Octet the_chained_P1                        ;
    Octet the_chained_P2                        ;
    Octet_String the_chained_data               ;

    Octet the_challenge_preservation_counter    ;
    Octet the_PIN_verification_counter          ;
	bool the_PIN_is_verified                    ;
};


extern "C" PIV_CARD_APPLICATION_DLL Application_Creator create_application ;

#endif  // __PIV_CARD_APPLICATION_H_INCLUDED__
