// Storage -- Persistent, reference-counted memory
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include <Storage.h>

#include <memory.h>
#include <sys/stat.h>

#include <Messages.h>

class Generic_Memory : public Memory
  {
  public:
    // Constructor is protected to disallow explicit instantiation
    Generic_Memory ( 
             // The size of the memory chunk this instance implements
             Memory::size_t size 
           ) 
      : the_size ( size ) ,
        the_reference_count ( 0 ) 
      { } ;


    // The type of a transaction level
    typedef unsigned short int transaction_level_t ;

    // The layout of the controlled memory area
    static const Memory::offset_t the_current_object_table_offset  = 0 * sizeof ( Memory::offset_t) ;
    static const Memory::offset_t the_heap_free_list_head_offset   = 1 * sizeof ( Memory::offset_t) ;

    static const Memory::size_t   granularity                      = 8 ; // TODO: compute

    static const Memory::offset_t the_free_storage_offset          = 2 * sizeof ( Memory::offset_t) < granularity 
                                                                      ? granularity
                                                                      : 2 * sizeof ( Memory::offset_t) ; 

    static const Octet           default_storage_value           = 0xFF;

    static
    void *
    pointer ( Memory::offset_t offset )
      {
        check_whether_initialized ( ) ;
        return reinterpret_cast < Octet * > ( the_storage ) + offset ; 
      } ;

    static
    Memory::offset_t 
    offset ( void * pointer )
      { 
        check_whether_initialized ( ) ;
        return static_cast < Memory::offset_t >
          ( reinterpret_cast < Octet * > ( pointer ) 
            - reinterpret_cast < Octet * > ( the_storage ) ) ; 
      } ;

    // The object table offset slot reference
    static
    Memory::offset_t & 
    current_object_table_offset ( )
      {
        return 
          * reinterpret_cast < Memory::offset_t * > 
              ( pointer ( the_current_object_table_offset ) );
      } ;

    // The heap free list head offset slot reference
    // The head of the heap free list
    static
    Memory::offset_t & 
    heap_free_list_head_offset ( )
      {
        return 
          * reinterpret_cast < Memory::offset_t * > 
              ( pointer ( the_heap_free_list_head_offset ) );
      } ;



    reference_count_t
    increment_reference_count ( )
      throw ( Reference_Count_Overflow_Exception )
      {
        reference_count_t the_new_reference_count = the_reference_count + 1 ;
        if ( the_new_reference_count < 0 )
          throw Reference_Count_Overflow_Exception ( ) ;
        return the_reference_count = the_new_reference_count ;
      } ;

    reference_count_t
    decrement_reference_count ( )
      throw ( Reference_Count_Underflow_Exception )
      {
        reference_count_t the_new_reference_count = the_reference_count - 1 ;
        if ( the_new_reference_count < 0 )
          throw Reference_Count_Underflow_Exception ( ) ;
        return the_reference_count = the_new_reference_count ;
      } ;


    static bool initialize ( size_t storage_size )
      {
        the_storage = malloc ( storage_size ) ;
        if ( 0 == the_storage )
          return false ;
        the_storage_size = storage_size ;
        return true ;
      } ;

    static void finalize ( )
      {
        if ( 0 != the_storage )
          {
            free ( the_storage ) ;
            the_storage = 0 ;
          }
        the_storage_size = 0 ;
      } ;

    static void check_whether_initialized ( )
      {
        if ( 0 == the_storage )
          throw Storage_Uninitialized_Exception ( ) ;
      } ;

    static bool
    load ( FILE * input , unsigned short & checksum ) ;

    static bool
    store ( FILE * output , unsigned short & checksum ) ;

    // The controlled memory area
    static void * the_storage ;
  
    // The controlled memory area size
    static size_t the_storage_size ;
  
    // The size of the memory chunk this instance implements
    size_t the_size ;

    // Is this a valid storage pointer?
    static bool is_storage_pointer ( void * pointer )
      {
        return 
          reinterpret_cast < std::size_t > ( the_storage ) 
            <= reinterpret_cast < std::size_t > ( pointer )
          && reinterpret_cast < std::size_t > ( pointer ) 
            <= reinterpret_cast < std::size_t > ( the_storage ) + the_storage_size ;
      } ;

    // The number of counted references to the memory chunk this instance implements
    reference_count_t the_reference_count ;

} ;

void * Generic_Memory::the_storage = 0 ;

Generic_Memory::size_t Generic_Memory::the_storage_size = 0 ;

// Read the heap from external persistent memory
bool
Generic_Memory::load ( FILE * input , unsigned short & checksum )
  {
    the_storage = 0 ;
    the_storage_size = 0 ;


    if ( 1 != fread ( & the_storage_size , sizeof ( the_storage_size ) , 1 , input ) )
      return false ;

    initialize ( the_storage_size ) ;

    if ( 1 != fread ( the_storage , the_storage_size , 1 , input ) )
      return false ;

    checksum ^= the_storage_size ;

    for ( unsigned short int * p = reinterpret_cast < unsigned short int * > ( the_storage ) ,
                             * pend = p + the_storage_size / sizeof ( unsigned short int ) ;
          p < pend ;
          p ++ )
      checksum ^= * p ;

    return true ;
  } ;


// Write the heap to external persistent memory
bool
Generic_Memory::store ( FILE * output , unsigned short int & checksum )
  {

    if ( 1 != fwrite ( & the_storage_size , sizeof ( the_storage_size ) , 1 , output ) )
      return false ;

    if ( 1 != fwrite ( the_storage , the_storage_size , 1 , output ) )
      return false ;

    checksum ^= the_storage_size ;

    for ( unsigned short int * p = reinterpret_cast < unsigned short int * > ( the_storage ) ,
                               * pend = p + the_storage_size / sizeof ( unsigned short int ) ;
          p < pend ;
          p ++ )
      checksum ^= * p ;

    return true ;
  } ;



class Free_Memory : public Generic_Memory
  {
  public:

    // Initialize the controlled heap
    static 
    void
    initialize (  ) ;

    Free_Memory ( 
           // The size of the memory chunk this instance implements
           Memory::size_t size , 
           // The link to the previous and next chunks on the heap free list
           Memory::offset_t xored_prev_and_next_offsets 
         )
      : Generic_Memory ( size )  ,
        the_xored_prev_and_next_offsets ( xored_prev_and_next_offsets )
      { } ;


    static
    Free_Memory *
    heap_pointer ( Memory::offset_t offset )
      { return reinterpret_cast < Free_Memory * > ( pointer ( offset ) ); } ;

    static 
    Memory::size_t
    round_up_to_granularity ( Memory::size_t size ) ;

    static
    Free_Memory *
    allocate ( Memory::size_t size )
    throw ( bad_alloc ) ; 
    
    static
    void
    deallocate ( Free_Memory * the_block_pointer ) ;

    static
    void
    dump ( ) ;

    Memory::offset_t 
    get_other_offset ( Memory::offset_t offset )
      { return the_xored_prev_and_next_offsets ^ offset ; } ;
    
    void 
    change_xored_offset ( Memory::offset_t old_offset , 
                          Memory::offset_t new_offset )
      { the_xored_prev_and_next_offsets ^= ( old_offset ^ new_offset ) ; } ;


    void 
    set_xored_offset ( Memory::offset_t prev_offset , 
                       Memory::offset_t next_offset )
      { the_xored_prev_and_next_offsets = ( prev_offset ^ next_offset ) ; } ;


    // The smallest free block size
    static const Memory::size_t minimum_heap_list_block_size ;

    // The link to the previous and next chunks on the heap free list
    Memory::offset_t the_xored_prev_and_next_offsets ;
  } ;



void
Free_Memory::initialize ( ) 
  { 
    memset ( Generic_Memory::the_storage , default_storage_value , Generic_Memory::the_storage_size ) ;

    Free_Memory * the_free_storage_pointer = heap_pointer ( the_free_storage_offset ) ;

    Memory::size_t the_free_storage_size =
      Generic_Memory::the_storage_size - the_free_storage_offset ;

    new ( the_free_storage_pointer ) Free_Memory ( the_free_storage_size , 0 ^ 0 ) ;

    heap_free_list_head_offset ( ) = the_free_storage_offset ;
  } ;


class Allocated_Memory : public Generic_Memory
  {
  public :
    Allocated_Memory ( 
                // The size of the memory chunk this instance implements
                Memory::size_t size , 
                // The transaction level in which this memory chunk is dirty
                transaction_level_t transaction_level
              )
      : Generic_Memory ( size )  ,
        the_transaction_level  ( transaction_level  ) ,
        the_object_table_index ( 0 )
      { } ;

    ~Allocated_Memory ( )
      { } ;

    void *
    object_pointer ( )
      { return this + 1 ; } ;

    const void *
    object_pointer ( )
    const
      { return this + 1 ; } ;

    Memory::offset_t
    body_offset ( )
      { return Generic_Memory::offset ( object_pointer ( ) ) ; } ;

    static 
    Allocated_Memory *
    allocate ( Memory::size_t size , transaction_level_t transaction_level )
    throw ( bad_alloc )
      {
        Allocated_Memory * the_object_pointer = 
          new ( Free_Memory::allocate ( size ) ) 
            Allocated_Memory ( size , transaction_level ) ;
        if ( 0 < size )
          memset ( the_object_pointer -> object_pointer ( ) , 0 , size ) ;
        return the_object_pointer ;
      } ;

    static 
    void
    free ( Allocated_Memory * the_allocated_object_pointer )
      {
        the_allocated_object_pointer -> ~Allocated_Memory ( );
        Free_Memory::deallocate ( reinterpret_cast < Free_Memory * > 
                             ( the_allocated_object_pointer ) );
      } ;

    Allocated_Memory *
    reallocate ( Memory::size_t the_new_size )
    throw ( bad_alloc )
      {
        // TODO:  This reallocation is too simple to be safe against
        //        tearing.  For ordinary object resizing, the new
        //        offset should be installed in the object table prior
        //        to freeing the old copy.  Similarly, for resizing the
        //        object table, the new object table should be installed
        //        prior to freeing the old one.
        Allocated_Memory * the_new_object_pointer = 
          allocate ( the_new_size , the_transaction_level ) ;
        Memory::size_t the_common_size =
          the_new_size < the_size ? the_new_size : the_size ;
        if ( 0 < the_common_size )
          memcpy ( the_new_object_pointer -> object_pointer ( ) ,  
                   object_pointer ( ) , 
                   the_common_size ) ;
        free ( this ) ;
        return the_new_object_pointer ;
      } ;

    static
    Allocated_Memory *
    from_object_pointer ( void * object_pointer )
      {
        return reinterpret_cast < Allocated_Memory * > ( object_pointer ) - 1 ;
      }

    static
    const Allocated_Memory *
    from_object_pointer ( const void * object_pointer )
      {
        return reinterpret_cast < const Allocated_Memory * > ( object_pointer ) - 1 ;
      }

    static
    Allocated_Memory *
    from_body_offset ( Memory::offset_t body_offset )
      { return from_object_pointer ( Generic_Memory::pointer ( body_offset ) ) ; }


    // The transaction level in which this memory chunk is dirty
    transaction_level_t the_transaction_level ;

    // The index of the memory chunk this instance implements
    object_table_index_t the_object_table_index ;

  } ; //  Allocated_Memory


inline Memory::size_t
Free_Memory::round_up_to_granularity ( Memory::size_t size )
{
  return ( (  size + sizeof ( Allocated_Memory) + granularity - 1 ) / granularity ) * granularity ;
} ;


Free_Memory *
Free_Memory::allocate ( Memory::size_t size )
throw ( bad_alloc )
  {
    if ( 2 < get_verbosity ( ) )
      {
        debug_message ( "\nallocate -- entering, size=%04X...\n", size) ;
        dump ( ) ;
      }

    Memory::offset_t the_prev_block_offset = 0 ;
    Memory::offset_t the_current_block_offset = heap_free_list_head_offset ( ) ;
    Memory::offset_t the_next_block_offset = 0 ;

    Free_Memory * the_current_block_pointer =
      heap_pointer ( the_current_block_offset ) ;

    Memory::size_t actual_size_including_header = 
      round_up_to_granularity ( size ) ;

    // Don't split a block unless the remainder is large enough to be linkable
    Memory::size_t minimum_splittable_size = 
      actual_size_including_header + minimum_heap_list_block_size ;

    for ( ; ; ) 
      {
        if ( 0 == the_current_block_offset  )
          {
            // TODO: try memory compaction
            throw bad_alloc ( "Memory::allocate failed" ) ;
          }

        Memory::size_t current_block_size = 
          the_current_block_pointer -> the_size ;
        the_next_block_offset    =
          the_current_block_pointer -> get_other_offset ( the_prev_block_offset ) ;
        // Take first fit
        if ( actual_size_including_header <= current_block_size )
          {
            // Found one -- should we split it?
            if ( minimum_splittable_size <= current_block_size )
              {
                // Split the block
                Memory::size_t new_block_size = 
                  current_block_size - actual_size_including_header ;
                the_current_block_pointer -> the_size = new_block_size ;
                the_current_block_offset = 
                  static_cast < Memory::size_t > 
                    ( the_current_block_offset + new_block_size ) ;
                the_current_block_pointer =
                  heap_pointer ( the_current_block_offset ) ;
              }
            else
              {
                // Take this block
                if ( 0 != the_next_block_offset )
                  {
                    Free_Memory * next_block_pointer =
                      heap_pointer ( the_next_block_offset ) ;
                    next_block_pointer -> change_xored_offset ( the_current_block_offset , 
                                                          the_prev_block_offset ) ;
                  }
                if ( 0 != the_prev_block_offset )
                  {
                    Free_Memory * prev_block_pointer =
                      heap_pointer ( the_prev_block_offset ) ;
                    prev_block_pointer -> change_xored_offset ( the_current_block_offset , 
                                                          the_next_block_offset ) ;
                  }
                else
                  {
                    // We are about to take the head block
                    heap_free_list_head_offset ( ) = the_next_block_offset ;
                  }
              }
            the_current_block_pointer -> the_size = size ;
            break ;
          }
        the_prev_block_offset     = the_current_block_offset ;
        the_current_block_offset  = the_next_block_offset ;
        the_current_block_pointer = heap_pointer ( the_current_block_offset ) ;
      }

    if ( 2 < get_verbosity ( ) )
      {
        debug_message ( "allocate -- ... exiting, allocated %04X\n" , the_current_block_offset) ;
        dump ( ) ;
      }

    return the_current_block_pointer ;
  } ;


void
Free_Memory::deallocate ( Free_Memory * the_block_pointer )
  {
    if ( 2 < get_verbosity ( ) )
      {
        debug_message ( "\ndeallocate -- entering, deallocating %08lX ...\n", the_block_pointer ) ;
        dump ( ) ;
      }

    Memory::offset_t the_block_offset = Generic_Memory::offset ( the_block_pointer ) ;
    Memory::size_t the_block_size =
      round_up_to_granularity ( the_block_pointer -> the_size ) ;
    memset ( the_block_pointer , default_storage_value , the_block_size ) ;
    the_block_pointer -> the_size = the_block_size ;

    Memory::offset_t the_prev_block_offset = 0 ;
    Memory::offset_t the_current_block_offset = heap_free_list_head_offset ( ) ;
    Memory::offset_t the_next_block_offset = 0 ;

    if ( 3 < get_verbosity ( ) )
      debug_message ( "\n0x%04X\n", the_block_offset ) ;
    while ( 0 != the_current_block_offset ) 
      {
        Free_Memory * the_current_block_pointer =
          heap_pointer ( the_current_block_offset ) ;
        the_next_block_offset = 
          the_current_block_pointer 
            -> get_other_offset ( the_prev_block_offset ) ;
        if ( the_block_offset < the_current_block_offset )
          break ;
        if ( 3 < get_verbosity ( ) )
          debug_message ( "0x%04X 0x%04X 0x%04X -- current=0x%08lX\n", 
                          the_prev_block_offset , 
                          the_current_block_offset , 
                          the_next_block_offset , 
                          the_current_block_pointer ) ;
        the_prev_block_offset = the_current_block_offset ;
        the_current_block_offset = the_next_block_offset ;
      } 
    if ( 3 < get_verbosity ( ) )
      debug_message ( "0x%04X 0x%04X 0x%04X\n\n", the_prev_block_offset , the_current_block_offset , the_next_block_offset ) ;


    // The block goes between the prev block and the current block 

    the_block_pointer -> 
      set_xored_offset ( the_prev_block_offset , the_current_block_offset ) ;

    // Consider the prev block
    if ( 0 != the_prev_block_offset )
      {
        Free_Memory * the_prev_block_pointer =
          heap_pointer ( the_prev_block_offset ) ;

        Memory::size_t the_prev_block_size = 
          the_prev_block_pointer -> the_size ;
        Memory::offset_t the_prev_block_end =
          the_prev_block_offset + the_prev_block_size ;

        if ( the_prev_block_end == the_block_offset ) 
          // Merge the prev block and the block
          {
            if ( 3 < get_verbosity ( ) )
              debug_message ( "***merge with prev: %04X+%04X\n", the_prev_block_offset , the_block_offset);
            memset ( the_block_pointer , default_storage_value , sizeof ( Free_Memory ) );
            the_block_size = 
              static_cast < Memory::size_t > 
                ( the_block_size + the_prev_block_size ) ;
            the_block_pointer = the_prev_block_pointer ;
            the_block_offset = the_prev_block_offset ;
            the_block_pointer -> the_size = the_block_size ;
          }
        else
          // Move the prev block link from the current block to the block
          the_prev_block_pointer 
            -> change_xored_offset ( the_current_block_offset , 
                                     the_block_offset ) ;
      }
    else
      {
        // This block becomes the free list head
        heap_free_list_head_offset ( ) = the_block_offset ;
      }

    // Consider the current block
    if ( 0 != the_current_block_offset )
      {
        
        Memory::offset_t the_block_end = 
          the_block_offset + the_block_size ;
        Free_Memory * the_current_block_pointer =
          heap_pointer ( the_current_block_offset ) ;

        if ( the_block_end == the_current_block_offset )
          // Merge the block with the current block 
          { 
            if ( 3 < get_verbosity ( ) )
              debug_message ( "***merge with current: %04X+%04X\n", 
                              the_block_offset , the_current_block_offset);
            the_block_size = 
              static_cast < Memory::size_t > 
                ( the_block_size 
                  + the_current_block_pointer -> the_size );
            the_block_pointer -> the_size = the_block_size ;
            Memory::offset_t the_next_block_offset =
              the_current_block_pointer 
                -> get_other_offset ( the_prev_block_offset ) ;
            memset ( the_current_block_pointer , default_storage_value , sizeof ( Free_Memory ) ) ;
            the_block_pointer 
              -> change_xored_offset ( the_current_block_offset , 
                                       the_next_block_offset ) ;
            if ( 0 != the_next_block_offset )
              heap_pointer ( the_next_block_offset ) 
                -> change_xored_offset ( the_current_block_offset , 
                                         the_block_offset ) ;
          }
        else
          // Move the current block link from the prev block to the block
          the_current_block_pointer 
            -> change_xored_offset ( the_prev_block_offset , 
                                     the_block_offset ) ;
      }

    if ( 2 < get_verbosity ( ) )
      {
        debug_message ( "deallocate -- ... exiting\n") ;
        dump_the_free_list ( ) ;
      }
  } ;


void
Free_Memory::dump ( )
throw ( const char * )
  {
    debug_message ( "Free_Memory Heap Block list:\n/---------------------------\\\n");
    Memory::offset_t the_prev_block_offset = 0 ;
    Memory::offset_t the_current_block_offset = heap_free_list_head_offset ( ) ;
    Memory::offset_t the_next_block_offset = 0 ;
    while ( 0 != the_current_block_offset )
      {
        Free_Memory * the_current_block_pointer = 
          reinterpret_cast < Free_Memory * >
            ( pointer ( the_current_block_offset )  ) ;
        the_next_block_offset = the_current_block_pointer -> get_other_offset ( the_prev_block_offset ) ;
        debug_message ( "%04X <- %04X -> %04X -- %04X\n", 
                        the_prev_block_offset , 
                        the_current_block_offset , 
                        the_next_block_offset ,
                        the_current_block_pointer -> the_size ) ;
        the_prev_block_offset = the_current_block_offset ;
        the_current_block_offset = the_next_block_offset ;
      }
    debug_message ( "\\---------------------------/\n");
  } ;


const Memory::size_t Free_Memory::minimum_heap_list_block_size =
  Free_Memory::round_up_to_granularity ( 0 )  ;




class Object_Table 
  {
  public :
    // A guess at the appropriate initial object table size
    // The object table will resize itself as needed, 
    // so the particular choice here doesn't matter too much, 
    // but it would obviously be most efficient if this were
    // exactly as big as it would ever need to be.
    // The previous size is recorded so that the sizes increase
    // as Fibonacci numbers.  See Knuth.

    typedef Memory::size_t                        size_t               ;
    typedef Generic_Memory::offset_t              offset_t             ;
    typedef Allocated_Memory::transaction_level_t transaction_level_t  ;
    typedef Memory::object_table_index_t          object_table_index_t ;


    static const object_table_index_t initial_object_table_previous_slots_count = 1 ;
    static const object_table_index_t initial_object_table_slots_count          = 1 ;

    static const transaction_level_t  initial_transaction_level        = 0 ;

    Object_Table ( transaction_level_t  transaction_level         ,
                   object_table_index_t previous_number_of_slots  ,
                   object_table_index_t number_of_slots           )
      : the_transaction_level        ( transaction_level ) ,
        the_previous_number_of_slots ( previous_number_of_slots  ) ,
        the_number_of_slots          ( number_of_slots  )
      { 
        memset ( this + 1 ,
                 0 ,
                 the_number_of_slots * sizeof ( Memory::offset_t ) ) ;
      } ;

    // Initialize the object table
    static
    void
    initialize ( ) ;

    Memory::offset_t &
    entry ( object_table_index_t index )
      {
        return 
          reinterpret_cast < Memory::offset_t * > ( this + 1 ) [ index - 1 ] ;
      }

    object_table_index_t
    find_free_entry_index ( )
      {
        Memory::offset_t * entries = ( & entry ( 1 ) ) - 1 ;
        for ( object_table_index_t i = 1 ; i <= the_number_of_slots ; i ++ )
          if ( 0 == entries [ i ] )
            return i ;
        return 0 ;
      }

    static 
    void *
    index_to_pointer ( object_table_index_t index )
      {
        if ( 0 == index  ) return 0 ;

        Memory::offset_t offset = 
          current_object_table_pointer ( ) -> entry ( index ) ;

        if ( 0 == offset ) return 0 ;

        return Generic_Memory::pointer ( offset ) ;
      }

    bool
    resize ( ) 
      {
        // Try to expand the current object table
        object_table_index_t the_new_number_of_slots =
          static_cast < object_table_index_t >
            ( the_number_of_slots + the_previous_number_of_slots ) ;

        Memory::size_t the_new_object_table_size =
          static_cast < Memory::size_t >
            ( sizeof ( Object_Table ) 
                + the_new_number_of_slots 
                    * sizeof ( Memory::offset_t ) ) ;
        try 
          {
            Object_Table * the_new_object_table_pointer =
              reinterpret_cast < Object_Table * >
                ( reinterpret_cast < Allocated_Memory * > 
                    ( Allocated_Memory::from_object_pointer ( this ) 
                      -> reallocate ( the_new_object_table_size ))
                  -> object_pointer ( ) ) ;
            the_new_object_table_pointer -> the_previous_number_of_slots =
              the_new_object_table_pointer -> the_number_of_slots ;
            the_new_object_table_pointer -> the_number_of_slots =
              the_new_number_of_slots ;

            Generic_Memory::current_object_table_offset ( ) = 
              Generic_Memory::offset ( the_new_object_table_pointer ) ;
            // TODO: only now is it safe to free the old object table
          }
        catch ( bad_alloc )
          {
            return false ;
          } ;
        return true ;
      } ;



    void
    dump ( ) ;

    static
    Object_Table *
    current_object_table_pointer ( )
      {
        return 
          reinterpret_cast < Object_Table * > 
            ( Generic_Memory::pointer ( Generic_Memory::current_object_table_offset ( ) ) ) ; 
      } ;

    transaction_level_t   the_transaction_level ;
    object_table_index_t  the_number_of_slots  ;
    object_table_index_t  the_previous_number_of_slots  ;
    Memory::offset_t      the_previous_transaction_level_object_table_offset ;
  } ; // Object_Table


void
Object_Table::initialize ( )
  {
    Memory::size_t the_object_table_size =
      static_cast < Memory::size_t >
        ( sizeof ( Object_Table ) 
            + initial_object_table_slots_count 
                * sizeof ( Memory::offset_t ) ) ;

    Object_Table * the_current_object_table_pointer =
      reinterpret_cast < Object_Table * >
        ( Allocated_Memory::allocate ( the_object_table_size , 0 ) 
            -> object_pointer ( ) ) ;
    
    new ( the_current_object_table_pointer ) 
      Object_Table ( initial_transaction_level , 
                     initial_object_table_slots_count , 
                     initial_object_table_previous_slots_count ) ;

    the_current_object_table_pointer -> 
      the_previous_transaction_level_object_table_offset = 0 ;

    Generic_Memory::current_object_table_offset ( ) = 
      Generic_Memory::offset ( the_current_object_table_pointer ) ;

  } ;



void
Object_Table::dump ( )
throw ( const char * )
  {
  if ( 0 == Generic_Memory::current_object_table_offset ( ) )
      throw "Object table not initialized" ;
  int number_of_occupied_slots = 0 ;

    debug_message ( "Object Table Entries[%d]: (%04X - 0x%08lX)\n"
                    "/---+------+------------\\\n",
                    the_number_of_slots, 
                    Generic_Memory::current_object_table_offset( ) ,
                    current_object_table_pointer ( ) );

    for ( object_table_index_t i = 1 ; i <= the_number_of_slots ; i ++ )
      {
        Memory::offset_t & this_entry = 
          current_object_table_pointer ( ) -> entry ( i ) ;
        if ( 0 != this_entry )
          {
            debug_message ( "%4d: %04X -> 0x%08lX\n", 
                            i , 
                            this_entry  , 
                            Generic_Memory::pointer ( this_entry ) ) ;
            number_of_occupied_slots ++ ;
          }
      }

    debug_message ( "\\---+------+------------/\n"
                    "%7d slot%s occupied.\n", 
                    number_of_occupied_slots, 
                    1==number_of_occupied_slots ? "" : "s");
  } ;


unsigned int Memory::the_verbosity = 0 ;

// Allocate and initialize the controlled heap
bool 
Memory::initialize ( Memory::size_t storage_size ) 
  { 
    if ( Generic_Memory::initialize ( storage_size ) )
      {
        Free_Memory::initialize ( ) ;
        Object_Table::initialize ( ) ;
        return true ;
      }
    return false ;
  } ;


// Deallocate the controlled heap
void 
Memory::finalize ( ) 
  { Generic_Memory::finalize ( ) ; } ;


// Read the heap from external persistent memory
bool
Memory::load ( FILE * input , unsigned short & checksum )
  { return Generic_Memory::load ( input , checksum ) ; } ;


// Write the heap to external persistent memory
bool
Memory::store ( FILE * output , unsigned short & checksum )
  { return Generic_Memory::store ( output , checksum ) ; } ;

void 
Memory::dump_the_free_list ( )
  { Free_Memory::dump ( ) ; } ;


void 
Memory::dump_the_current_object_table ( )
  { Object_Table::current_object_table_pointer ( ) -> dump ( ); } ;


void *
Memory::operator new ( std::size_t  _Count )
throw ( bad_alloc )
  {
    Object_Table * the_current_object_table_pointer = 
      Object_Table::current_object_table_pointer ( ) ;
    object_table_index_t the_object_index = 
      the_current_object_table_pointer -> find_free_entry_index ( ) ;

    // If the table is full, try to resize it
    if ( 0 == the_object_index && the_current_object_table_pointer -> resize ( ) )
      {
        the_current_object_table_pointer = 
          Object_Table::current_object_table_pointer ( ) ;
        the_object_index = 
          the_current_object_table_pointer -> find_free_entry_index ( ) ;
      }

    // If there is [still] no empty slot, fail
    if ( 0 == the_object_index )
      {
        if ( 1 < get_verbosity ( ) )
          debug_message ("Memory::operator new(%3d) failed.\n" , _Count ) ;
        throw 
          bad_alloc ( "Memory::operator new: could not find free object table entry" ) ;
      }

    Allocated_Memory * the_allocated_memory_pointer = 
      Allocated_Memory::allocate ( static_cast < Memory::size_t > ( _Count ) , 
                            the_current_object_table_pointer 
                              -> the_transaction_level ) ;

    the_current_object_table_pointer -> entry ( the_object_index ) 
      = the_allocated_memory_pointer -> body_offset ( ) ;

    the_allocated_memory_pointer -> the_object_table_index = the_object_index ;

    if ( 1 < get_verbosity ( ) )
      debug_message ("Memory::operator new(%3d) -> %3d (0x%08lx)\n" ,
                     _Count ,
                     the_object_index ,
                     the_allocated_memory_pointer -> object_pointer ( ) ) ;

    return the_allocated_memory_pointer -> object_pointer ( ) ;
  } ;


void 
Memory::operator delete ( void * _Ptr )
  {
    if ( 0 != _Ptr )
      {
        Allocated_Memory * the_object_pointer =
          Allocated_Memory::from_object_pointer ( _Ptr ) ;
        object_table_index_t the_object_index = 
          the_object_pointer -> the_object_table_index ;
        if ( 0 != the_object_index )
          {
            if ( 1 < get_verbosity ( ) )
              debug_message ("Memory::operator delete(0x%08lx) - %3d\n" ,
                            the_object_pointer ,
                            the_object_index ) ;
            Object_Table * the_current_object_table_pointer = 
              Object_Table::current_object_table_pointer ( ) ;
            Generic_Memory::Memory::offset_t & the_object_entry =
              the_current_object_table_pointer -> entry ( the_object_index ) ;
            Allocated_Memory::free ( the_object_pointer ) ;
            the_object_entry = 0 ;        
          }
      }
  } ;



void *
Memory::from_object_pointer ( void * object_pointer )
  {
    return reinterpret_cast < Allocated_Memory * > ( object_pointer ) - 1 ;
  } ;


void *
Memory::index_to_pointer ( object_table_index_t object_table_index  )
  {
    return Object_Table::index_to_pointer ( object_table_index ) ;
  } ;



void indexed_pointer_base::add_reference_to_referenced_object ( void * the_object_pointer )
  {
    if ( the_object_pointer != 0 ) 
      {
        // Check to make sure this object pointer points into the persisent object storage
      if ( ! Allocated_Memory::is_storage_pointer ( the_object_pointer ) )
          {
            debug_message ( "indexed_pointer_base::add_reference_to_referenced_object: not persistent storage pointer:  pointer=0x%08lX\n" ,
                            the_object_pointer ) ;
            throw "indexed_pointer_base::add_reference_to_referenced_object: not persistent storage pointer" ;
          }
        
        Allocated_Memory * the_memory_block_pointer =
          reinterpret_cast < Allocated_Memory * > 
            ( Memory::from_object_pointer ( the_object_pointer ) ) ;
        the_object_table_index =
          the_memory_block_pointer -> the_object_table_index ;
        if ( the_object_pointer != object_pointer ( ) )
          {
            debug_message ( "indexed_pointer_base::add_reference_to_referenced_object: broken invariant: pointer ( index ( pointer))!=pointer\n"
                            "pointer=0x%08lX index=%04X pointer ( index)=0x%08lX\n" ,
                            the_object_pointer ,
                            the_object_table_index ,
                            object_pointer ( ) ) ;
            throw "indexed_pointer_base::add_reference_to_referenced_object: broken invariant: pointer ( index ( pointer))!=pointer" ;
          }
        the_memory_block_pointer -> increment_reference_count ( ) ;
      }
    else
      the_object_table_index = 0 ;
  } ;


bool
indexed_pointer_base::remove_reference_to_referenced_object ( void * the_object_pointer )
  {
    the_object_table_index = 0 ;
    if ( the_object_pointer != 0 )
      {
        if ( reinterpret_cast < Allocated_Memory * > 
                   ( Memory::from_object_pointer ( the_object_pointer ) )
                 -> decrement_reference_count ( ) 
               == 0 )
        return true ;
      }
    return false ;
  } ;

void indexed_pointer_base::deallocate_referenced_object ( void * the_object_pointer )
  { Memory::operator delete ( the_object_pointer ) ; } ;


bool
indexed_pointer_base::store ( FILE * the_persistent_storage_output_file , unsigned short & checksum ) const
  {
    bool result =
      1 == fwrite ( & the_object_table_index , 
                    sizeof ( the_object_table_index ) , 
                    1 , 
                    the_persistent_storage_output_file ) ;

    if ( result ) checksum ^= the_object_table_index ;

    return result ;
  }

bool 
indexed_pointer_base::load ( FILE * the_persistent_storage_input_file , unsigned short & checksum )
  {
    bool result =
      1 == fread ( & the_object_table_index , 
                    sizeof ( the_object_table_index ) , 
                    1 , 
                    the_persistent_storage_input_file ) ;

    if ( result ) checksum ^= the_object_table_index ;

    return result ;
  }
