// TLP_APDU_Client -- TLP224 Listener
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#include <SocketIO.h>
#include <clains.h>
#include <sw.h>

#include "TLP_APDU_Client.h"


const char TLP_APDU_Client::default_hostname [ ] = "localhost" ; 

#define TLP_WINSOCK_2_0_NOT_SUPPORTED  ( WSABASEERR - 1 )
#define TLP_NO_APDUSERVER              ( WSABASEERR - 2 )
#define TLP_COMMAND_TOO_LONG           ( WSABASEERR - 3 )
#define TLP_UNKNOWN_CARD_NUMBER        ( WSABASEERR - 4 )
#define TLP_OUTPUT_BUFFER_OVERFLOW     ( WSABASEERR - 5 )
#define TLP_INPUT_BUFFER_UNDERFLOW     ( WSABASEERR - 6 )
#define TLP_ERROR_SENDING_PACKET       ( WSABASEERR - 7 )
#define TLP_ERROR_RECEIVING_PACKET     ( WSABASEERR - 8 )

#define MAXIMUM_DATA_BYTES_PER_RESPONSE     248

static void print_TLP_APDU_Client_error(long result)
{
   error_message ( "TLP APDU Client error %d\n", result ) ;
}


long TLP_APDU_Client::initialize ( int verbosity )
{
   long result ;

   the_current_logical_card_identifier = (unsigned short)-1;

   result = APDU_Client::initialize ( verbosity ) ;
   if ( 0 != result )
   {
      // We depend on the APDU_Server to generate
      // its own error messages during failed
      // initialization.
      goto cleanup;
   }


   result = SocketIO_initialize ( ) ;
   if ( 0 != result )
     goto cleanup;

  if ( the_verbosity )
    log_message("TLP_APDU_Client: Listening for connection.\n" ) ;

  result = 
    SocketIO_initialize_listener ( the_port, 
                                   & the_listening_socket , 
                                   & the_connection_socket ) ;
  if ( 0 != result )
    {
      // We depend on SocketIO to generate
      // its own error messages during failed
      // initialization.
      return result ;
    }

  if ( the_verbosity )
    log_message("TLP_APDU_Client: Connected.\n" ) ;
  
  if ( the_verbosity )
    log_message ( "TLP_APDU_Client initialized.\n");

  the_saved_response_data . set_APDU_length ( 0 ) ;

  return result ;

cleanup:
  error_message ( "TLP_APDU_Client initialization failed.\n");
  return result ;
}


long TLP_APDU_Client::run ( void )
{
  static char comment [ 4096 ] ;
  static char filepath [ _MAX_PATH ] ;
  long result = 0 ;
  unsigned char command ;
  unsigned char OSB ;
  int bytesSent = 0 ; 
  int bytesToBeSent = 0 ;
  int bytesTotal = 0 ;
  unsigned char logical_card_number_low_byte  ;
  unsigned char logical_card_number_high_byte ;
  int i ;
  int length = 0 ;
  unsigned char SW1, SW2 ;

  CAPDU_Packet the_CAPDU ;
  RAPDU_Packet the_RAPDU ;

  // Process commands
  for ( ; ; )
    {
      if ( the_verbosity )
        log_message("TLP_APDU_Client: Listening for command packet.\n" ) ;

      if ( ! the_TLP224 . receive_TLP224_packet( the_connection_socket , command , length ) )
        {
          // Has the client just terminated?
          if ( INVALID_SOCKET == the_connection_socket )
            {
              if ( the_verbosity )
                log_message("TLP_APDU_Client: Connection broken.\n" ) ;
              break ;
            }
          result = TLP_ERROR_RECEIVING_PACKET ;
          print_TLP_APDU_Client_error(result);
          break ;
        }

      if ( the_verbosity )
        log_message("TLP_APDU_Client: Command packet received.\n" ) ;

      OSB = TLP224::OSB_OK ;
      switch ( command )
        {
          case TLP224::TLPC_TERMINATE                 :
              return result;

            case TLP224::TLPC_POWER_DOWN                :
              the_APDU_server . power_down ( ) ;
              return result ;

            case TLP224::TLPC_POWER_UP                  :
              the_APDU_server . power_up ( ) ;
              {
                unsigned int the_ATR_length ;
                unsigned char the_ATR [ 32 ] ;  // TODO: 32
                the_APDU_server . get_ATR ( the_ATR , the_ATR_length ) ;
                OSB = 0 ;
                if ( ! ( the_TLP224 . put_byte_into_output_buffer( 0x28 ) 
                        && the_TLP224 . put_byte_into_output_buffer( 0x02 ) 
                        && the_TLP224 . put_byte_into_output_buffer( ( unsigned char ) the_ATR_length ) ) )
                  {
                    // No room in the buffer
                    OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                    return result ;
                  }
                for ( i = 0 ; i < ( int ) the_ATR_length ; i ++ )
                  if ( ! ( the_TLP224 . put_byte_into_output_buffer( the_ATR [ i ] ) ))
                    {
                      // No room in the buffer
                      OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                      return result ;
                    }
                bytesToBeSent = static_cast < int > ( the_ATR_length + 1 ) ;
              }
              break ;

            case TLP224::TLPC_OPEN_LOG_FILE             :
              memset(filepath  , 0 , sizeof(filepath)  );
              for ( i = 0 ; i < length ; i ++ )
              {
                  if ( 0 == the_TLP224 
                              . get_byte_from_input_buffer
                                  ( * reinterpret_cast < unsigned char * > 
                                          ( & filepath [ i ] ) ) )
                  {
                    // Not as many bytes as promised
                    OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                    return result ;
                  }
              }
              if ( the_verbosity )
              {
                  log_message("TLP_APDU_Client: Opening log file \"%s\" ....\n" , filepath) ;
              }
              length = 0 ;
              if ( 0 != the_APDU_server . open_log_file ( filepath ) )
              {
                  // Server couldn't open file
                  OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                  return result ;
              }
              if ( the_verbosity )
              {
                  log_message("TLP_APDU_Client: Log file \"%s\" opened\n" , filepath) ;
              }
              break ;

            case TLP224::TLPC_FLUSH_LOG_FILE            :
              length = 0 ;
              if ( 0 != the_APDU_server . flush_log_file ( ) )
              {
                  // Server couldn't flush file
                  OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                  return result ;
              }
              if ( the_verbosity )
              {
                  log_message("TLP_APDU_Client: Log file flushed\n") ;
              }
              break ;

            case TLP224::TLPC_CLOSE_LOG_FILE            :
              length = 0 ;
              if ( 0 != the_APDU_server . close_log_file ( ) )
              {
                  // Server couldn't flush file
                  OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                  return result ;
              }
              if ( the_verbosity )
              {
                  log_message("TLP_APDU_Client: Log file flushed\n") ;
              }
              break ;

            case TLP224::TLPC_LOG_FILE_COMMENT          :
              if ( sizeof(comment) < length )
              {
                  // Comment too long
                  OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                  return result ;
              }
              memset(comment  , 0 , sizeof(comment)  );
              for ( i = 0 ; i < length ; i ++ )
              {
                  if ( 0 == the_TLP224 . get_byte_from_input_buffer( * reinterpret_cast < unsigned char * > ( & comment [ i ] ) ) )
                  {
                    // Not as many bytes as promised
                    OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                    return result ;
                  }
              }
              if ( the_verbosity )
              {
                  log_message("TLP_APDU_Client: Writing log file comment \"%s\" ....\n" , comment) ;
              }
              length = 0 ;
              if ( 0 != the_APDU_server .log_file_comment ( comment ) )
              {
                  // Server couldn't open file
                  OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                  return result ;
              }
              if ( the_verbosity )
              {
                  log_message("TLP_APDU_Client: Log file comment \"%s\" written\n" , comment) ;
              }
              break ;


            case TLP224::TLPC_SELECT_CARD_NUMBER   :
              // Select a different card, if necessary
              the_TLP224 . get_byte_from_input_buffer( logical_card_number_high_byte );
              the_TLP224 . get_byte_from_input_buffer( logical_card_number_low_byte );
              the_current_logical_card_identifier = MAKEWORD( logical_card_number_low_byte , logical_card_number_high_byte) ;
              if ( the_verbosity )
              {
                  log_message("TLP_APDU_Client: Card Number set to 0x%04X\n", the_current_logical_card_identifier ) ;
              }
              break ;


            case TLP224::TLPC_LOAD_CARD_IMAGE           :
              // Unknown functionality
              break ;

            case TLP224::TLPC_STORE_CARD_IMAGE          :
              // Unknown functionality
              break ;

            case TLP224::TLPC_EXCHANGE_APDU_CARD_0      :
            case TLP224::TLPC_OUTPUT_TO_MICROPROCESSOR  :
            case TLP224::TLPC_INPUT_FROM_MICROPROCESSOR :

            the_CAPDU . set_APDU_length ( length ) ;
            the_CAPDU . set_logical_card_identifier ( the_current_logical_card_identifier ) ;
            for ( i = 0 ; i < length ; i ++ )
            {
                unsigned char b ;
                if ( 0 == the_TLP224 . get_byte_from_input_buffer( b ) )
                {
                    // Missing length byte
                    OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                    return result ;
                }
                the_CAPDU . put_APDU_octet ( i , b ) ;
            }

            if ( 0 < the_verbosity )
            {
                log_message("TLP_APDU_Client: Command APDU: Length %d CardNum 0x%04X cAPDU:",
                        the_CAPDU . get_APDU_length ( ),
                        the_CAPDU . get_logical_card_identifier ( )) ;
                for ( i = 0 ; i < static_cast < int > ( the_CAPDU . get_APDU_length ( ) ) ; i ++ )
                {
                    log_message(" %02X", the_CAPDU . get_APDU_octet ( i ));
                }
                log_message("\n");
            }
                
            // Respond to GET RESPONSE
            if (  CAPDU_Packet::CLAINS_GET_RESPONSE == the_CAPDU . get_APDU_u16 ( CAPDU_Packet::OFFSET_CLA ) )
            {
                // Respond to GET RESPONSE
                if ( bytesSent <  bytesTotal )
                {
                    if ( bytesTotal - bytesSent < MAXIMUM_DATA_BYTES_PER_RESPONSE )
                    {
                        bytesToBeSent = bytesTotal - bytesSent ;
                        SW1 = 0x90 ;
                        SW2 = 0x00 ;
                    }
                    else
                    {
                        bytesToBeSent = MAXIMUM_DATA_BYTES_PER_RESPONSE ;
                        if ( bytesTotal - ( bytesSent + bytesToBeSent ) < MAXIMUM_DATA_BYTES_PER_RESPONSE )
                        {
                            SW1 = 0x61 ;
                            SW2 = bytesTotal - ( bytesSent + bytesToBeSent ) ;
                        }
                        else
                        {
                            SW1 = 0x61 ;
                            SW2 = MAXIMUM_DATA_BYTES_PER_RESPONSE ;
                        }
                    }
                }
                else
                {
                    the_RAPDU . set_status_only ( RAPDU_Packet::SW_INS_INVALID ) ;
                    result = 1 ;
                }
            }
            else
            {
                // Exchange APDU with server
                the_saved_response_data . set_APDU_length ( 0 ) ;
                result = the_APDU_server . exchange_APDUs ( the_CAPDU , the_RAPDU ) ;
                if ( 0 != result )
                {
                    print_TLP_APDU_Client_error(result);
                    return result ;
                }
                bytesSent = 0;
                bytesToBeSent = ( int )the_RAPDU . get_APDU_length ( ) ;
                SW2 = the_RAPDU . get_APDU_octet ( --bytesToBeSent ) ;
                SW1 = the_RAPDU . get_APDU_octet ( --bytesToBeSent ) ;
                bytesTotal = bytesToBeSent ;
            }

            // Definitely use GET RESPONSE if the R-APDU is longer than MAXIMUM_BYTES_PER_RESPONSE
            if ( bytesToBeSent > MAXIMUM_DATA_BYTES_PER_RESPONSE )
            {
#ifdef EMPTY_61XX_NOT_NEEDED
                bytesToBeSent = MAXIMUM_DATA_BYTES_PER_RESPONSE ;
                if ( bytesTotal - ( bytesSent + bytesToBeSent ) < MAXIMUM_DATA_BYTES_PER_RESPONSE )
                {
                    SW1 = 0x61 ;
                    SW2 = bytesTotal - ( bytesSent + bytesToBeSent ) ;
                }
                else
                {
                    SW1 = 0x61 ;
                    SW2 = MAXIMUM_DATA_BYTES_PER_RESPONSE ;
                }
#endif
                SW1 = 0x61 ;
                SW2 = MAXIMUM_DATA_BYTES_PER_RESPONSE ;
                bytesToBeSent = 0 ;
            }
            // It's the usual case but possibly use GET RESPONSE for Case 4 APDUs
            else if ( use_GET_RESPONSE
                   && CAPDU_Packet::OFFSET_CDATA < the_CAPDU . get_APDU_length ( ) 
                   && sizeof ( unsigned short ) < the_RAPDU . get_APDU_length ( ) )
            {
                SW1 = 0x61 ;
                SW2 = bytesToBeSent ;
                bytesToBeSent = 0 ;
            }

            if ( 0 < the_verbosity )
            {
                log_message("TLP_APDU_Client: Response APDU: Length %d CardNum 0x%04X rAPDU:",
                        the_RAPDU . get_APDU_length ( ),
                        the_RAPDU . get_logical_card_identifier ( )) ;
                for ( i = 0 ; i < static_cast < int > ( the_RAPDU . get_APDU_length ( ) ) ; i ++ )
                {
                    log_message(" %02X", the_RAPDU . get_APDU_octet ( i ));
                }
                log_message("\n");
            }

            // Put the out-going bytes into the TLP-224 buffer
            for ( i = 0 ; i < bytesToBeSent ; i ++ )
            {
                if ( ! the_TLP224 . put_byte_into_output_buffer( the_RAPDU . get_APDU_octet ( bytesSent + i ) ) )
                {
                    OSB = TLP224::OSB_OTHER_THAN_JUST_OK ;
                    return result ;
                }
            }
            bytesSent += bytesToBeSent ;
            the_TLP224 . put_byte_into_output_buffer( SW1 ) ;
            the_TLP224 . put_byte_into_output_buffer( SW2 ) ; 
            break;

            default:
            error_message ("Unrecognized command 0x%02X in packet\n", command);
            break;
        }

        if ( ! the_TLP224 . send_TLP224_packet ( the_connection_socket , OSB , bytesToBeSent + sizeof( SW1 ) + sizeof( SW1 ) ) )
        {
            result = TLP_ERROR_SENDING_PACKET ;
            print_TLP_APDU_Client_error(result);
            break ;
        }
        if ( the_verbosity )
        {
            log_message("TLP_APDU_Client: Response packet sent.\n" ) ;
        }
    }
    return ( result ) ;
}


long TLP_APDU_Client::finalize ( void )
{
   unsigned long int result = 0 ;

  SocketIO_finalize_listener ( & the_listening_socket , & the_connection_socket ) ;
  
  result = APDU_Client :: finalize ( ) ;
  // We depend on the APDU_Server to
  // generate its own error messages if
  // finalization fails.

  SocketIO_finalize ( ) ;
  // We depend on SocketIO to
  // generate its own error messages if
  // finalization fails.

  if ( the_verbosity )
    log_message("TLP_APDU_Client: Finalized.\n") ;

  return ( result ) ;
}
