// APDU_Packet - Data Structures for Network-Word-Order APDU Wrapper Packets
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__APDU_PACKET_H_INCLUDED__)
#define __APDU_PACKET_H_INCLUDED__

// An APDUPacket looks like this
// size_t  Length                   ==> Total length of this structure
// size_t  APDULength;              ==> Number of payload bytes in APDU
// unsigned short                   ==> Logical Card Identifier


#include <stdlib.h>
#include <Messages.h>

#include <string>
using namespace std;

#define FRAME_SIZE 0x32c8  //13000 byte max frame size

class APDU_Packet 
{
public:
  const static size_t header_length = 
    sizeof ( unsigned long )        // total object size   (4)
      + sizeof ( unsigned long  )   // APDU payload length (4)
      + sizeof ( unsigned short ) ; // logical card id     (2)
  const static size_t offset_cla  = header_length;
  const static size_t offset_ins  = offset_cla + 1;
  const static size_t offset_p1   = offset_ins + 1;
  const static size_t offset_p2   = offset_p1 + 1;
  const static size_t offset_p3   = offset_p2 + 1;
  const static size_t offset_data = offset_p3 + 1;

  APDU_Packet ( size_t frame_size = FRAME_SIZE )
    : the_frame_size ( frame_size )
    { the_frame = reinterpret_cast < unsigned char * > ( malloc ( frame_size ) ) ; } ;

  APDU_Packet ( const APDU_Packet & other )
    : the_frame_size ( other . the_frame_size )
    { 
      the_frame = reinterpret_cast < unsigned char * > ( malloc ( the_frame_size ) ) ; 
      memcpy ( the_frame , other . the_frame , the_frame_size ) ;
    } ;

  APDU_Packet & operator = ( const APDU_Packet & other )
    { 
      the_frame_size = other . the_frame_size ;
      the_frame = reinterpret_cast < unsigned char * > ( malloc ( the_frame_size ) ) ; 
      memcpy ( the_frame , other . the_frame , the_frame_size ) ;
      return * this ;
    } ;

  ~APDU_Packet ( )
    { free ( the_frame ) ; } ;

  APDU_Packet ( const unsigned char * octets , size_t length , size_t frame_size = 512 )
    : the_frame_size ( frame_size )
    { 
      the_frame = reinterpret_cast < unsigned char * > ( malloc ( frame_size ) ) ;
      set_APDU ( octets , length ) ; 
    } ;
  
  size_t
  max_APDU_size ( ) const
    {  return the_frame_size - header_length ; } ;


  unsigned char * 
  APDU ( ) 
    { return the_frame + header_length ; } ; 

  const unsigned char * 
  APDU ( ) const
    { return the_frame + header_length ; } ; 

  const unsigned char * set_APDU ( const unsigned char * new_APDU , size_t new_APDU_length )
    {
      set_APDU_length ( new_APDU_length ) ;
      memcpy( APDU ( ) , new_APDU , new_APDU_length ) ;
      return new_APDU  ;
    }

  // Number of payload bytes in APDU
  size_t get_APDU_length ( ) const
    {
      size_t ret_value = ( (size_t) ( the_frame [ 7 ] ) << 24 )
                          + ( (size_t) ( the_frame [ 6 ] ) << 16 )
                          + ( (size_t) ( the_frame [ 5 ] ) <<  8 )
                          + ( (size_t) ( the_frame [ 4 ] )       );
      return ret_value ;
    };

	size_t set_APDU_length ( size_t new_length )
    {
      debug_message ( "APDU_Packet.set_APDU_length ( %d )\n" , static_cast < unsigned int > ( new_length ) ) ;
      
      the_frame [ 7 ] = (unsigned char)( ( new_length & 0xff000000 ) >> 24 ) ;
      the_frame [ 6 ] = (unsigned char)( ( new_length & 0x00ff0000 ) >> 16 ) ;
      the_frame [ 5 ] = (unsigned char)( ( new_length & 0x0000ff00 ) >>  8 ) ;
      the_frame [ 4 ] = (unsigned char)( ( new_length & 0x000000ff )       ) ;

      return 0 ;
    }
  
  // addressed card
  unsigned short get_logical_card_identifier ( ) const
    {
      unsigned short  ret_value = 
                                ( (unsigned short) ( the_frame [ 9 ] ) <<  8 ) +
                                ( (unsigned short) ( the_frame [ 8 ] )       );
        
      return ret_value ;
    } ;

  // addressed card
  unsigned short set_logical_card_identifier ( unsigned short new_card_id )
    {
      the_frame [ 9 ] = ( ( new_card_id & 0xff00 ) >> 8 ) ;
      the_frame [ 8 ] = ( ( new_card_id & 0x00ff )      ) ;    

      return 0 ;
    } ;

  unsigned char get_APDU_octet ( const size_t index ) const
    { return the_frame [ header_length + index ] ; } ;


  unsigned char put_APDU_octet ( const size_t index, const unsigned char value )
    { return the_frame [ header_length + index ] = value ; } ;


  const unsigned short int 
  get_APDU_u16 ( const size_t index ) const // throw ( ... ) 
    {
      return get_APDU_octet ( index ) * 256 + get_APDU_octet ( index + 1 ) ;
    } ;

  unsigned short int 
  put_APDU_u16 ( const size_t index, unsigned short int u16 )  // throw ( ... ) 
    {
      put_APDU_octet ( index     , static_cast <unsigned char> ( u16 / 256 ) ) ;
      put_APDU_octet ( index + 1 , static_cast <unsigned char> ( u16 % 256 ) ) ;
      return u16 ;
    } ;

  static const size_t APDU_packet_frame_size = 512 ;


  string 
  spell ( void ) const ;

  void 
  log   (const char * prefix) const ;
  void 
  debug (const char * prefix) const ;
  void 
  error (const char * prefix) const ;


  unsigned char * the_frame ;
  size_t the_frame_size ;
} ;


typedef unsigned short int CLAINS ;

class CAPDU_Packet : public APDU_Packet
{
public:
  static const int OFFSET_CLA   = 0 ;
  static const int OFFSET_INS   = 1 ;
  static const int OFFSET_P1    = 2 ;
  static const int OFFSET_P2    = 3 ;
  static const int OFFSET_P3    = 4 ;
  static const int OFFSET_CDATA = 5 ;

//
// CLA and INS definitions for all implemented APDUs
//

// macro for coding CAPDUs
#define CLA_INS(clains) HIBYTE(clains), LOBYTE(clains)

//
// ISO Commands
//
  static const unsigned short CLAINS_EXTERNAL_AUTHENTICATE              = 0x0082 ;
  static const unsigned short CLAINS_INTERNAL_AUTHENTICATE              = 0x0088 ;
  static const unsigned short CLAINS_GET_CHALLENGE                      = 0x0084 ;

  static const unsigned short CLAINS_GENERAL_AUTHENTICATE               = 0x0087 ; //SP-800-73
  static const unsigned short CLAINS_GENERAL_AUTHENTICATE_CHAIN         = 0x1087 ; //SP-800-73

  static const unsigned short CLAINS_MANAGE_SECURITY_ENVIRONMENT        = 0x0022 ;
  static const unsigned short CLAINS_PERFORM_SECURITY_OPERATION         = 0x002A ;

  static const unsigned short CLAINS_SELECT                             = 0x00A4 ; //SP-800-73
  static const unsigned short CLAINS_GET_RESPONSE                       = 0x00C0 ;
  static const unsigned short CLAINS_ENVELOPE                           = 0x00C2 ;
  static const unsigned short CLAINS_MANAGE_CHANNEL                     = 0x0070 ;

  static const unsigned short CLAINS_VERIFY                             = 0x0020 ; //SP-800-73
  static const unsigned short CLAINS_CHANGE_REFERENCE_DATA              = 0x0024 ; //SP-800-73
  static const unsigned short CLAINS_RESET_RETRY_COUNTER                = 0x002C ; //SP-800-73

  static const unsigned short CLAINS_CREATE_FILE                        = 0x00E0 ;
  static const unsigned short CLAINS_DELETE_FILE                        = 0x00E4 ;

  static const unsigned short CLAINS_TERMINATE_CARD_USAGE               = 0x00FE ;
  static const unsigned short CLAINS_TERMINATE_DF                       = 0x00E6 ;
  static const unsigned short CLAINS_TERMINATE_EF                       = 0x00E8 ;

  static const unsigned short CLAINS_READ_BINARY                        = 0x00B0 ;
  static const unsigned short CLAINS_UPDATE_BINARY                      = 0x00D6 ;
  static const unsigned short CLAINS_ERASE_BINARY                       = 0x000E ;
  static const unsigned short CLAINS_SEARCH_BINARY                      = 0x00A0 ;

  static const unsigned short CLAINS_READ_RECORD                        = 0x00B2 ;
  static const unsigned short CLAINS_UPDATE_RECORD                      = 0x00DC ;
  static const unsigned short CLAINS_APPEND_RECORD                      = 0x00E2 ;
  static const unsigned short CLAINS_SEARCH_RECORD                      = 0x00A2 ;
  static const unsigned short CLAINS_ERASE_RECORD                       = 0x000C ;

  static const unsigned short CLAINS_DEACTIVATE_FILE                    = 0x0004 ;
  static const unsigned short CLAINS_ACTIVATE_FILE                      = 0x0044 ;

  static const unsigned short CLAINS_DISABLE_VERIFICATION_REQUIREMENT   = 0x0026 ;
  static const unsigned short CLAINS_ENABLE_VERIFICATION_REQUIREMENT    = 0x0028 ;

  static const unsigned short CLAINS_GET_DATA                           = 0x00CB ; //SP-800-73
  static const unsigned short CLAINS_PUT_DATA                           = 0x00DB ; //SP-800-73
  static const unsigned short CLAINS_PUT_DATA_CHAIN                     = 0x10DB ; //SP-800-73

  static const unsigned short CLAINS_GENERATE_ASYMMETRIC_KEY_PAIR       = 0x0047 ; //SP-800-73
  static const unsigned short CLAINS_GENERATE_ASYMMETRIC_KEY_PAIR_CHAIN = 0x1047 ; //SP-800-73

//
// GlobalPlatform APDUs
//
  static const unsigned short CLAINS_LOAD                               = 0x80E8 ;
  static const unsigned short CLAINS_INSTALL                            = 0x80E6 ;

//
// ETSI Commands
//
  static const unsigned short CLAINS_FETCH                              = 0x8012 ;
  static const unsigned short CLAINS_INCREASE                           = 0x8032 ;
  static const unsigned short CLAINS_STATUS                             = 0x00F2 ;
  static const unsigned short CLAINS_TERMINALRESPONSE                   = 0x8014 ;
//
// Invalid Command
//
  static const unsigned short CLAINS_INVALID                            = 0xFFFF ;

//
// CLA and INS Flags
//
  static const unsigned short CLAINS_SECURE_MESSAGING_FLAG              = 0x0800 ;
  static const unsigned short CLAINS_SECURE_MESSAGING_MASK              = 0xEC00 ;
  static const unsigned short CLAINS_BER_TLV                            = 0x0001 ;
  static const unsigned short CLAINS_COMMAND_CHAINING                   = 0x1000 ;

  unsigned char get_CLA ( )  const
    { return get_APDU_octet ( OFFSET_CLA ) ; } ;

  unsigned char get_INS ( )  const
    { return get_APDU_octet ( OFFSET_INS ) ; } ;

  CLAINS get_CLAINS ( )  const
    { return get_APDU_u16 ( OFFSET_CLA ) ; } ;

  unsigned char get_P1 ( )  const
    { return get_APDU_octet ( OFFSET_P1 ) ; } ;

  unsigned char get_P2 ( )  const
    { return get_APDU_octet ( OFFSET_P2 ) ; } ;

  unsigned short int get_P1P2 ( )  const
    { return get_APDU_u16 ( OFFSET_P1 ) ; } ;

  unsigned char get_P3 ( )  const
    { return get_APDU_octet ( OFFSET_P3 ) ; } ;

  unsigned char get_Le ( )  const
    { return get_APDU_octet ( get_APDU_length( ) - 1 ) ; } ;

  unsigned char * payload ( size_t offset = 0 )
    { return APDU ( ) + OFFSET_CDATA + offset ; } ;

  const unsigned char * payload ( size_t offset = 0 ) const
    { return APDU ( ) + OFFSET_CDATA + offset ; } ;

  unsigned char get_payload_octet ( size_t offset = 0 )  const
    { return get_APDU_octet ( offset + OFFSET_CDATA  ) ; } ;

  unsigned short int get_payload_u16 ( size_t offset = 0 )  const
    { return get_APDU_u16 ( offset + OFFSET_CDATA  ) ; } ;

} ;


class RAPDU_Packet : public APDU_Packet
  {
  public:
    typedef unsigned short int Status_Word ;

    /* Normal Processing */
    static const Status_Word SW_OK                                   = 0x9000 ; //SP-800-73
    static const Status_Word SW_EXTRA_INFORMATION                    = 0x9100 ;
    static const Status_Word SW_LENGTH                               = 0x6100 ; //SP-800-73
    static const Status_Word SW_RESPONSE                             = 0x6C00 ;

    /* Warnings */
    static const Status_Word SW_NO_INFORMATION                       = 0x6200 ;
    static const Status_Word SW_CANNOT_OPEN_LOGICAL_CHANNEL          = 0x6200 ;
    static const Status_Word SW_CORRUPTED_RETURN                     = 0x6281 ;
    static const Status_Word SW_END_OF_FILE                          = 0x6282 ; //SP-800-73
    static const Status_Word SW_FILE_INVALIDATED                     = 0x6283 ;
    static const Status_Word SW_FCI_NOT_CORRECT                      = 0x6284 ;
    static const Status_Word SW_SELECTED_FILE_TERMINATED             = 0x6285 ;
    static const Status_Word SW_VERIFICATION_FAILED                  = 0x6300 ; //SP-800-73
    static const Status_Word SW_RETRIES_REMAINING                    = 0x63C0 ; //SP-800-73
    static const Status_Word SW_0_RETRIES_REMAINING                  = 0x63C0 ; //SP-800-73

    /* Execution Errors */
    static const Status_Word SW_MEMORY_UNCHANGED                     = 0x6400 ;
    static const Status_Word SW_MEMORY_CHANGED                       = 0x6500 ;
    static const Status_Word SW_MEMORY_PROBLEM                       = 0x6581 ;
    static const Status_Word SW_SE_NOT_MODIFIED                      = 0x6600 ;

    /* Checking Errors */
    static const Status_Word SW_WRONG_LENGTH                         = 0x6700 ;
    static const Status_Word SW_WRONG_P1_P2                          = 0x6B00 ;
    static const Status_Word SW_INS_INVALID                          = 0x6D00 ;
    static const Status_Word SW_CLA_INVALID                          = 0x6E00 ;
    static const Status_Word SW_TECHNICAL_PROBLEM                    = 0x6F00 ;
    static const Status_Word SW_COMMAND_PROBLEM                      = 0x6F00 ;
    static const Status_Word SW_UNKNOWN                              = 0x6F00 ;

    /* Functions in CLA Not Supported */
    static const Status_Word SW_CLA_NO_INFORMATION                   = 0x6800 ; //SP-800-73
    static const Status_Word SW_LOGICAL_CHANNEL                      = 0x6881 ;
    static const Status_Word SW_SECURE_MESSAGING                     = 0x6882 ;
    static const Status_Word SW_LAST_COMMAND_OF_CHAIN_EXPECTED       = 0x6883 ;
    static const Status_Word SW_COMMAND_CHAINING_NOT_SUPPORTED       = 0x6884 ;

    /* Command Not Allowed */
    static const Status_Word SW_BAD_COMMAND_NO_INFORMATION           = 0x6900 ;
    static const Status_Word SW_INCOMPATIBLE_FILE_STRUCTURE          = 0x6981 ;
    static const Status_Word SW_SECURITY_STATUS_NOT_SATISFIED        = 0x6982 ; //SP-800-73
    static const Status_Word SW_AUTHENTICATION_METHOD_BLOCKED        = 0x6983 ; //SP-800-73
    static const Status_Word SW_REFERENCED_DATA_INVALIDATED          = 0x6984 ;
    static const Status_Word SW_CONDITIONS_OF_USE_NOT_SATISFIED      = 0x6985 ; //SP-800-73
    static const Status_Word SW_COMMAND_NOT_ALLOWED                  = 0x6986 ;
    static const Status_Word SW_NO_CURRENT_FILE                      = 0x6986 ;
    static const Status_Word SW_SM_DATA_OBJECT_MISSING               = 0x6987 ;
    static const Status_Word SW_SM_DATA_OBJECT_INCORRECT             = 0x6988 ;

    /* Wrong Parameters */
    static const Status_Word SW_WRONG_PARAMETERS_NO_INFORMATION      = 0x6A00 ;
    static const Status_Word SW_WRONG_PARAMETERS_IN_DATA_FIELD       = 0x6A80 ; //SP-800-73
    static const Status_Word SW_FUNCTION_NOT_SUPPORTED               = 0x6A81 ; //SP-800-73
    static const Status_Word SW_DATA_OBJECT_NOT_FOUND                = 0x6A82 ; //SP-800-73
    static const Status_Word SW_FILE_NOT_FOUND                       = 0x6A82 ;
    static const Status_Word SW_APP_NOT_FOUND                        = 0x6A82 ;
    static const Status_Word SW_RECORD_NOT_FOUND                     = 0x6A83 ;
    static const Status_Word SW_NOT_ENOUGH_SPACE                     = 0x6A84 ; //SP-800-73
    static const Status_Word SW_NC_INCONSISTENT_WITH_TLV             = 0x6A85 ;
    static const Status_Word SW_INCORRECT_P1_P2                      = 0x6A86 ; //SP-800-73
    static const Status_Word SW_LC_DOES_NOT_MATCH_P1_P2              = 0x6A87 ;
    static const Status_Word SW_NC_DOES_NOT_MATCH_P1_P2              = 0x6A87 ;
    static const Status_Word SW_REFERENCED_DATA_NOT_FOUND            = 0x6A88 ; //SP-800-73
    static const Status_Word SW_FILE_ALREADY_EXISTS                  = 0x6A89 ;
    static const Status_Word SW_DF_NAME_ALREADY_EXISTS               = 0x6A8A ;

    /* Application Errors */
    static const Status_Word SW_INCREASE_AT_MAX_VALUE                = 0x9850 ;
    static const Status_Word SW_AUTHENTICATION_ERROR                 = 0x9862 ;

    unsigned char * payload ( )
      { return APDU ( ) ; } ;

    void set_status ( Status_Word sw )
      { 
        size_t length = get_APDU_length ( ) ;
        set_APDU_length ( length + 2 ) ;
        put_APDU_u16 ( length , sw ) ;
      } ;

    void set_status_only ( Status_Word sw )
      {
        set_APDU_length ( 0 ) ;
        set_status ( sw ) ;
      }

    Status_Word get_status ( )
      { 
        size_t length = get_APDU_length ( ) ;
        return get_APDU_u16 ( length - 2 ) ;
      } ;
  } ;

#endif // !defined(__APDU_PACKET_H_INCLUDED__)
