// Storage -- Persistent, reference-counted memory
//
// COPYRIGHT
// ---------
// The software was developed by the National Institute of Standards and 
// Technology (NIST), an agency of the Federal Government.  Pursuant to 
// Title 15 United States Code Section 105, works of NIST are not subject
// to copyright protection in the United States and are considered to be
// in the public domain.  As a result, a formal license is not needed to
// use the software.
// 
// The Government Smart Card Reference Implementation software is provided
// by NIST as a service and is expressly provided "AS IS."  
// NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
// INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY.
// NIST does not warrant or make any representations regarding the use of
// the software or the results thereof, including but not limited to the
// correctness, accuracy, reliability or usefulness of the software.
// 
// Permission to use this software is contingent upon your acceptance of 
// the terms of this agreement and upon your providing appropriate
// acknowledgments of NIST's ownership of the software.

#if !defined(__STORAGE_H_INCLUDED__)
#define __STORAGE_H_INCLUDED__

#include <memory>
using namespace std;

#include <stdio.h>

#if defined(_MSC_VER )
  #if defined(_DEBUG)
    #include <crtdbg.h>
  #endif
  #define snprintf _snprintf
  // Quoting the Microsoft Visual C++ .NET helps:
  //    A function is declared using exception specification, 
  //    which Visual C++ accepts but does not implement. 
  //    Code with exception specifications that are ignored 
  //    during compilation may need to be recompiled and linked 
  //    to be reused in future versions supporting exception specifications.
  //
  //    You can avoid this warning by using the warning pragma:

  #pragma warning ( disable : 4290 4512 )
  #if defined(STORAGE_EXPORTS)
    #define STORAGE_DLL _declspec(dllexport)
  #else //  defined ( STORAGE_EXPORTS )
    #define STORAGE_DLL _declspec(dllimport)
  #endif  // defined ( STORAGE_EXPORTS )
#else
   #define STORAGE_DLL
#endif

// Typedefs

// The type of the unit of the contents of storage
typedef unsigned char      Octet ;


// The memory chunk class
class STORAGE_DLL Memory
  {
  public:
    // The type of an object table index
    typedef unsigned int object_table_index_t ;


    // The type of a reference count
    typedef           int reference_count_t ;

    // The type of a memory chunk size
    typedef unsigned  int size_t ;

    // The type of the offsets into controlled memory 
    typedef unsigned int offset_t ;

    // The type of the differences of offsets into controlled memory 
    typedef          int offset_difference_t ;

    // Execption throw on reference count overflow
    class Storage_Uninitialized_Exception { } ;

    // Execption throw on reference count overflow
    class Reference_Count_Overflow_Exception { } ;

    // Execption throw on reference count underflow
    class Reference_Count_Underflow_Exception { } ;

    // Allocate and initialize the controlled heap
    static bool
    initialize ( size_t storage_size ) ;

    // Deallocate the controlled heap
    static void
    finalize ( ) ;


    // Read the heap from external persistent memory
    // Update checksum
    static 
    bool
    load ( FILE * input , unsigned short & checksum ) ;

    // Write the heap to external persistent memory
    // Update checksum
    static 
    bool
    store ( FILE * output , unsigned short & checksum ) ;


    // Dump the current object table
    static 
    void
    dump_the_current_object_table ( ) ;


    // Dump the free list
    static 
    void
    dump_the_free_list ( ) ;

    // Throttle debugging Verbosity
    static 
    unsigned int
    get_verbosity ( )
      { return the_verbosity ; } ;

    static 
    unsigned int
    set_verbosity ( unsigned int verbosity )
      { return the_verbosity = verbosity ; } ;

    // Allocate a memory chunk of a specified size

    void *
    operator new ( std::size_t _Count )
    throw ( bad_alloc ) ;

    void *
    operator new ( std::size_t , void * _Ptr )
      { return _Ptr ; }  ;

    void 
    operator delete ( void * _Ptr ) ;

    void 
    operator delete ( void * , void * ) 
      { } ;

    static void * index_to_pointer ( object_table_index_t object_table_index ) ;

    static void * from_object_pointer ( void * object_pointer ) ;

  protected :

    // Constructor is protected to disallow explicit instantiation
    Memory ( ) { } ;


    // Verbosity throttle
    static unsigned int the_verbosity ;
  } ; // Memory



class STORAGE_DLL indexed_pointer_base
  {
  public :
    indexed_pointer_base ( )  : the_object_table_index ( 0 )  { } ;

    bool operator == ( const indexed_pointer_base & _Other ) const
      { return the_object_table_index == _Other . the_object_table_index ; } ;

    bool operator != ( const indexed_pointer_base & _Other ) const
      { return the_object_table_index != _Other . the_object_table_index ; } ;

    bool operator < ( const indexed_pointer_base & _Other ) const
      { return the_object_table_index < _Other . the_object_table_index ; } ;

    Memory::object_table_index_t get_index ( ) const
      { return the_object_table_index ; } ;

    void * object_pointer ( ) const
      { return Memory::index_to_pointer ( the_object_table_index ) ; } ;

    bool
    store ( FILE * the_persistent_storage_output_file , unsigned short & checksum ) const ;

    bool 
    load ( FILE * the_persistent_storage_input_file , unsigned short & checksum ) ;

  protected :

    void add_reference_to_referenced_object ( void * the_object_pointer ) ;
    bool remove_reference_to_referenced_object ( void * the_object_pointer ) ;
    void deallocate_referenced_object ( void * the_object_pointer ) ;

    Memory::object_table_index_t the_object_table_index ;
  } ;


template < typename _Ty > class indexed_pointer : public indexed_pointer_base
  {
  public:
    indexed_pointer ( )
      { } ;

    indexed_pointer ( const indexed_pointer & _Other )
      { attach ( _Other . object_pointer ( ) ) ; } ;

    ~indexed_pointer ( ) 
      { detach ( object_pointer ( ) ) ; } ;

    static indexed_pointer null ( ) { return indexed_pointer ( ) ; } ;

    indexed_pointer & operator = ( const indexed_pointer & _Other ) 
      {
        detach (          object_pointer ( ) ) ;
        attach ( _Other . object_pointer ( ) ) ;
        return * this ;
      } ;

    indexed_pointer & operator = ( _Ty * p ) 
      {
        detach ( object_pointer ( ) ) ;
        attach ( p ) ;
        return * this ;
      } ;

    const _Ty * operator -> ( ) const
      { return reinterpret_cast < const _Ty * > ( object_pointer ( ) ) ; } ;

          _Ty * operator -> ( )       
      { return reinterpret_cast <       _Ty * > ( object_pointer ( ) ) ; } ;

    const _Ty & operator * ( ) const { return * operator -> ( ) ; } ;

          _Ty & operator * ( )       { return * operator -> ( ) ; } ;


    static indexed_pointer 
    object_pointer_to_indexed_pointer ( _Ty * _Ptr )
      { indexed_pointer p ; p . attach ( _Ptr ) ;  return p ; } ;

  protected:
    void attach ( void * the_object_pointer )
      { add_reference_to_referenced_object ( the_object_pointer  ) ;} ;


    void detach ( void * the_object_pointer )
      {
        if ( remove_reference_to_referenced_object ( the_object_pointer ) )
          {
            // simulate "delete this"
            reinterpret_cast < _Ty * > ( the_object_pointer ) -> ~_Ty ( );
            deallocate_referenced_object ( the_object_pointer ) ;
          }
      } ;

  } ;



template < > class indexed_pointer < void > : public indexed_pointer_base
  {
  public:
    indexed_pointer < void > ( ) { } ;

    indexed_pointer < void > ( const indexed_pointer < void > & _Other )
      { attach ( _Other . object_pointer ( ) ) ; } ;

    ~indexed_pointer < void > ( ) 
      { detach ( object_pointer ( ) ) ; } ;

    static indexed_pointer < void > null ( ) { return indexed_pointer < void > ( ) ; } ;

    indexed_pointer < void > & operator = ( const indexed_pointer <void > & _Other )
      {
        detach (          object_pointer ( ) ) ;
        attach ( _Other . object_pointer ( ) ) ;
        return * this ;
      } ;

    indexed_pointer < void > & operator = ( void * p ) 
      {
        detach ( object_pointer ( ) ) ;
        attach ( p ) ;
        return * this ;
      } ;

    const void * operator -> ( ) const
      { return reinterpret_cast < const void * > ( object_pointer ( ) ) ; } ;

          void * operator ->  ( )       
      { return reinterpret_cast <       void * > ( object_pointer ( ) ) ; } ;

    static indexed_pointer < void >
    object_pointer_to_indexed_pointer ( void * _Ptr )
      { indexed_pointer <void > p ; p . attach ( _Ptr ) ;  return p ; } ;

    void attach ( void * the_object_pointer )
      { add_reference_to_referenced_object ( the_object_pointer ) ; } ;


    void detach ( void * the_object_pointer )
      {
        if ( remove_reference_to_referenced_object ( the_object_pointer ) )
          deallocate_referenced_object ( the_object_pointer ) ;
      } ;

  } ;

          
    // TEMPLATE FUNCTION _Persistent_Allocate
template<class _Ty> inline
	indexed_pointer < _Ty > _Persistent_Allocate ( Memory::size_t _Count, indexed_pointer < _Ty > )
	{	// allocate storage for _Count elements of type _Ty
    return 
      indexed_pointer < _Ty > :: object_pointer_to_indexed_pointer 
        ( (  _Ty * ) Memory::operator new ( _Count * sizeof ( _Ty ) ) ) ;
	}

		// TEMPLATE FUNCTION _Persistent_Construct
template<class _T1,
	class _T2> inline
	void _Persistent_Construct ( _T1 * _Ptr, const _T2& _Val)
	{	// construct object at _Ptr with value _Val
    new ( _Ptr ) _T1 ( _Val);
	}

		// TEMPLATE FUNCTION _Persistent_Destroy
template<class _Ty> inline
	void _Persistent_Destroy ( _Ty * _Ptr)
	{	// destroy object at _Ptr
  	_DESTRUCTOR ( _Ty, _Ptr);
	}

template<> inline
	void _Persistent_Destroy ( char * )
	{	// destroy a char ( do nothing)
	}

template<> inline
	void _Persistent_Destroy ( wchar_t * )
	{	// destroy a wchar_t ( do nothing)
	}

		// TEMPLATE CLASS _Persistent_Allocator_base
template<class _Ty>
	struct _Persistent_Allocator_base
	{	// base class for generic allocators
	typedef _Ty value_type;
	};

		// TEMPLATE CLASS _Persistent_Allocator_base<const _Ty>
template<class _Ty>
	struct _Persistent_Allocator_base<const _Ty>
	{	// base class for generic allocators for const _Ty
	typedef _Ty value_type;
	};

		// TEMPLATE CLASS persistent_allocator
template<class _Ty>
	class persistent_allocator
		: public _Persistent_Allocator_base<_Ty>
	{	// generic persistent_allocator for objects of class _Ty
public:
	typedef _Persistent_Allocator_base<_Ty> _Mybase;
	typedef typename _Mybase::value_type value_type;


	typedef indexed_pointer < value_type > pointer;
	typedef value_type& reference;
	typedef const indexed_pointer < value_type > const_pointer;
	typedef const value_type& const_reference;

	typedef Memory::size_t size_type;
	typedef Memory::offset_difference_t difference_type;

	template<class _Other>
		struct rebind
		{	// convert an persistent_allocator<_Ty> to an persistent_allocator <_Other>
		typedef persistent_allocator<_Other> other;
		};

	pointer address ( reference _Val) const
		{	// return address of mutable _Val
		return ( &_Val);
		}

	const_pointer address ( const_reference _Val) const
		{	// return address of nonmutable _Val
		return ( &_Val);
		}

	persistent_allocator ( )
		{	// construct default persistent_allocator ( do nothing)
		}

	persistent_allocator ( const persistent_allocator<_Ty>&)
		{	// construct by copying ( do nothing)
		}

	template<class _Other>
		persistent_allocator ( const persistent_allocator<_Other>&)
		{	// construct from a related persistent_allocator ( do nothing)
		}

	template<class _Other>
		persistent_allocator<_Ty>& operator= ( const persistent_allocator<_Other>&)
		{	// assign from a related persistent_allocator ( do nothing)
		return ( *this);
		}

	void deallocate ( pointer _Ptr, size_type)
		{	// deallocate object at _Ptr, ignore size
    Memory::operator delete ( & * _Ptr);
		}

	pointer allocate ( size_type _Count)
		{	// allocate array of _Count elements
		return ( _Persistent_Allocate ( _Count, pointer ( )));
		}

	pointer allocate ( size_type _Count, const indexed_pointer < void > )
		{	// allocate array of _Count elements, ignore hint
		return ( allocate ( _Count));
		}

	void construct ( _Ty * _Ptr, const _Ty& _Val)
		{	// construct object at _Ptr with value _Val
		_Persistent_Construct ( _Ptr, _Val);
		}

	void destroy ( _Ty * _Ptr)
		{	// destroy object at _Ptr
		_Persistent_Destroy ( _Ptr);
		}

	Memory::size_t max_size ( ) const
		{	// estimate maximum array size
		Memory::size_t _Count = ( Memory::size_t) ( -1) / sizeof ( _Ty);
		return ( 0 < _Count ? _Count : 1);
		}
	};

		// persistent_allocator TEMPLATE OPERATORS
template<class _Ty,
	class _Other> inline
	bool operator== ( const persistent_allocator<_Ty>&, const persistent_allocator<_Other>&)
	{	// test for persistent_allocator equality ( always true)
	return ( true);
	}

template<class _Ty,
	class _Other> inline
	bool operator!= ( const persistent_allocator<_Ty>&, const persistent_allocator<_Other>&)
	{	// test for persistent_allocator inequality ( always false)
	return ( false);
	}

		// CLASS persistent_allocator<void>
template<> class _CRTIMP2 persistent_allocator<void>
	{	// generic persistent_allocator for type void
public:
	typedef void _Ty;
	typedef indexed_pointer < _Ty > pointer;
	typedef const indexed_pointer < _Ty > const_pointer;
	typedef _Ty value_type;

	template<class _Other>
		struct rebind
		{	// convert an persistent_allocator<void> to an persistent_allocator <_Other>
		typedef persistent_allocator<_Other> other;
		};

	persistent_allocator ( )
		{	// construct default persistent_allocator ( do nothing)
		}

	persistent_allocator ( const persistent_allocator<_Ty>&)
		{	// construct by copying ( do nothing)
		}

	template<class _Other>
		persistent_allocator ( const persistent_allocator<_Other>&)
		{	// construct from related persistent_allocator ( do nothing)
		}

	template<class _Other>
		persistent_allocator<_Ty>& operator= ( const persistent_allocator<_Other>&)
		{	// assign from a related persistent_allocator ( do nothing)
		return ( *this);
		}
	};

		// TEMPLATE FUNCTION _Persistent_Destroy_range
template<class _Ty,
	class _Alloc> inline
	void _Persistent_Destroy_range ( indexed_pointer < _Ty > _First, indexed_pointer < _Ty > _Last, _Alloc& _Al)
	{	// destroy [_First, _Last)
	_Persistent_Destroy_range ( _First, _Last, _Al, _Ptr_cat ( _First, _Last));
	}

template<class _Ty,
	class _Alloc> inline
	void _Persistent_Destroy_range ( indexed_pointer < _Ty > _First, indexed_pointer < _Ty > _Last, _Alloc& _Al,
		_Nonscalar_ptr_iterator_tag)
	{	// destroy [_First, _Last), arbitrary type
	for ( ; _First != _Last; ++_First)
		_Al.destroy ( _First);
	}

template<class _Ty,
	class _Alloc> inline
	void _Persistent_Destroy_range ( indexed_pointer < _Ty > _First, indexed_pointer < _Ty > _Last, _Alloc& _Al,
		_Scalar_ptr_iterator_tag)
	{	// destroy [_First, _Last), scalar type ( do nothing)
	}


template < typename _storage_unit > class storage : public Memory
  {
  public:
    typedef indexed_pointer < storage < _storage_unit > > Pointer ;

    typedef persistent_allocator < storage < _storage_unit > > Allocator ;

    class Index_Exception
      {
      public:
        Index_Exception ( offset_t index ) { } ;
      } ;

    storage < _storage_unit > ( size_t _Count )
      : the_size ( _Count )
      { } ;

    static Pointer alloc ( size_t _Count ) 
      { 
        Pointer p =
          Pointer::object_pointer_to_indexed_pointer 
            ( reinterpret_cast < storage * >
                ( Memory::operator new 
                  ( sizeof ( storage ) 
                      + _Count * sizeof ( _storage_unit ) ) ) ) ;
        p -> the_size = _Count ;
        return p ;
      } ;

    static void free ( Pointer & p )
      { 
        Allocator ( ) . deallocate ( p , 0 ) ; 
        p = 0 ;
      } ;

    _storage_unit * body ( )
      { return reinterpret_cast < _storage_unit * > ( this + 1 ) ; } ;

    const _storage_unit * body ( ) const
      { return reinterpret_cast < const _storage_unit * > ( this + 1 ) ; } ;

    static void set ( Pointer & dst , const _storage_unit c , size_t size )
      { dst -> set ( c , size  ) ; } ;

    void
    set ( const _storage_unit c , size_t size )
      {
        check_index ( size - 1 ) ;
        _storage_unit * the_storage_units = body ( ) ;
        for ( offset_t i = 0 ; i < size ; i ++ )
          the_storage_units [ i ] = c ;
      } ;

    static
    int
    cmp ( const Pointer & a , offset_t a_start ,
          const Pointer & b , offset_t b_start ,
          size_t length )
      {
        a -> check_index ( a_start ) ;
        a -> check_index ( a_start + length - 1 ) ;
        b -> check_index ( b_start ) ;
        b -> check_index ( b_start + length - 1 ) ;
        const _storage_unit * the_a_units = a -> body ( ) + a_start  ;
        const _storage_unit * the_b_units = b -> body ( ) + b_start ;
        for ( offset_t i = 0 ; i < length ; i ++ )
          {
            const _storage_unit a_i = the_a_units [ i ] ;
            const _storage_unit b_i = the_b_units [ i ] ;
            if ( a_i < b_i ) return -1 ;
            else if ( b_i < a_i ) return +1 ;
          }
        return 0 ;
      } ;

    static
    Pointer &
    cpy (       Pointer & dst , offset_t dst_start ,
          const Pointer & src , offset_t src_start ,
          size_t length )
      {
        dst -> check_index ( dst_start ) ;
        dst -> check_index ( dst_start + length - 1 ) ;
        src -> check_index ( src_start ) ;
        src -> check_index ( src_start + length - 1 ) ;
        _storage_unit * the_dst_units = dst -> body ( ) + dst_start  ;
        const _storage_unit * the_src_units = src -> body ( ) + src_start ;
        for ( offset_t i = 0 ; i < length ; i ++ )
          the_dst_units [ i ] = the_src_units [ i ] ;
        return dst ;
      } ;

    static
    Pointer &
    cpy ( Pointer & dst , offset_t dst_start ,
          const _storage_unit * the_src_units ,
          size_t length )
      {
        dst -> check_index ( dst_start ) ;
        dst -> check_index ( dst_start + length - 1 ) ;
        _storage_unit * the_dst_units = dst -> body ( ) + dst_start  ;
        for ( offset_t i = 0 ; i < length ; i ++ )
          the_dst_units [ i ] = the_src_units [ i ] ;
        return dst ;
      } ;

    static
    Pointer &
    cpy (       Pointer & dst , offset_t dst_start ,
          const Pointer & src , offset_t src_start )
      {
        src -> check_index ( src_start ) ;
        const _storage_unit * the_src_units = src -> body ( ) + src_start ;
        size_t length = src -> the_size - src_start ;
        return cpy ( dst , dst_start , the_src_units , length ) ;
      } ;

    static
    Pointer &
    cpy (       Pointer & dst , offset_t dst_start ,
          const Pointer & src )
      {
        const _storage_unit * the_src_units = src -> body ( ) ;
        size_t length = src -> the_size ;
        return cpy ( dst , dst_start , the_src_units , length ) ;
      } ;

    static
    Pointer
    dup ( const _storage_unit * units , size_t size )
      {
        Pointer p = alloc ( size ) ;
        _storage_unit * the_storage_units = p -> body ( ) ;
        for ( offset_t i = 0 ; i < size ; i ++ )
          the_storage_units [ i ] = units [ i ] ;
        return p ;
      } ;

    _storage_unit
    get ( offset_t index )
    const
      {
        check_index ( index ) ;
        return body ( ) [ index ] ; 
      } ;

    _storage_unit
    put ( offset_t index , const _storage_unit c )
      { 
        check_index ( index ) ;
        return body ( ) [ index ] = c ; 
      } ;

    bool
    index_is_in_bounds ( offset_t index )
    const
      { return 0 <= index && index < the_size ; } ;

    size_t get_size ( ) const { return the_size ; } ;


  private :
    void
    check_index ( offset_t index )
    const
    throw ( Index_Exception )
      { if ( ! index_is_in_bounds ( index ) ) throw Index_Exception ( index ) ; } ;

    size_t the_size ;
  } ;


// A simple class that persistently stores some Octets
class STORAGE_DLL Octets : public storage < Octet >
  { 
  public:
    Octets ( size_t _Count = 0 )  : storage < Octet > ( _Count )  { } ;
  };


/*****
template < typename _pointer_type > class pointer_storage : public storage < _pointer_type >
  {
  public:
    pointer_storage < _pointer_type > ( size_t _Count )
      : storage < _pointer_type > ( _Count )
      { } ;

    ~pointer_storage < _pointer_type > ( )
      : ~ storage < _pointer_type > ( )
      {
        for ( size_t index = 0 ; index < the_size ; index ++ )
          body ( ) [ index ] = (_pointer_type)0 ;
      } ;
  } ;
*****/



// A simple class that is persistent
class STORAGE_DLL Object : public Memory
  {
  public:
    typedef indexed_pointer < Object > Pointer ;

    typedef persistent_allocator < Object > Allocator ;

    Object ( )
      { } ;

    static Pointer alloc ( ) 
      { return Allocator ( ) . allocate ( 1 ) ; } ;

    static void free ( Pointer & p )
      { Allocator ( ) . deallocate ( p , 0 ) ; } ;

  } ;


#endif  //  !defined(__STORAGE_H_INCLUDED__)
